/*---------------------------------------------------------------------------*/

/*!  \file  adi.h
     \brief  ADI subroutines for the field evolver.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#ifndef INCLUDE_ADI_H
#define INCLUDE_ADI_H

/*---------------------------------------------------------------------------*/
/*!  \brief  calculate the source for 1D ADI.
     \param  nx  number of division in x axis
     \param  dx2  square of the size of the mesh in x direction
     \param  h  factor for ADI (3 * \a dx^2 / \a omega / \a b^2) 
     \param  omega  acceleration factor
     \param  src  source for ADI
     \param  psi  psi field
     \param  mu  chemical potential
     \param  mu_ext  external force field  */
void calculate_adi_source_1d(int nx,double dx2,
                             double h,double omega,
                             double *src,double *psi,
                             double *mu,double *mu_ext);

/*!  \brief  calculate the source for 2D ADI.
     \param  nx  number of division in x axis
     \param  ny  number of division in y axis
     \param  dx2  square of the size of the mesh in x direction
     \param  dy2  square of the size of the mesh in y direction
     \param  h  factor for ADI (6 * \a dx^2 / \a omega / \a b^2) 
     \param  omega  acceleration factor
     \param  src  source for ADI
     \param  psi  psi field
     \param  mu  chemical potential
     \param  mu_ext  external force field  */
void calculate_adi_source_2d_x(int nx,int ny,
                               double dx2,double dy2,
                               double h,double omega,
                               double *src,double *psi,
                               double *mu,double *mu_ext);
void calculate_adi_source_2d_y(int nx,int ny,
                               double dx2,double dy2,
                               double h,double omega,
                               double *src,double *psi,
                               double *mu,double *mu_ext);

/*!  \brief  calculate the source for 3D ADI.
     \param  nx  number of division in x axis
     \param  ny  number of division in y axis
     \param  nz  number of division in y axis
     \param  dx2  square of the size of the mesh in x direction
     \param  dy2  square of the size of the mesh in y direction
     \param  dz2  square of the size of the mesh in y direction
     \param  h  factor for ADI (9 * \a dx^2 / \a omega / \a b^2) 
     \param  omega  acceleration factor
     \param  src  source for ADI
     \param  psi  psi field
     \param  mu  chemical potential
     \param  mu_ext  external force field  */
void calculate_adi_source_3d_x(int nx,int ny,int nz,
                               double dx2,double dy2,double dz2,
                               double h,double omega,
                               double *src,double *psi,
                               double *mu,double *mu_ext);
void calculate_adi_source_3d_y(int nx,int ny,int nz,
                               double dx2,double dy2,double dz2,
                               double h,double omega,
                               double *src,double *psi,
                               double *mu,double *mu_ext);
void calculate_adi_source_3d_z(int nx,int ny,int nz,
                               double dx2,double dy2,double dz2,
                               double h,double omega,
                               double *src,double *psi,
                               double *mu,double *mu_ext);

/*!  \brief  calculate the source for 1D ADI for dynamics.
     \param  nx  number of division in x axis
     \param  dx2  square of the size of the mesh in x direction
     \param  c  factor for ADI
     \param  omega  acceleration factor
     \param  phi  source for ADI (density field)
     \param  psi  psi field
     \param  L_psi  Laplacian of \a psi
     \param  mu_tot  chemical potential
     \param  L_mu_tot  Laplacian of \a mu_tot  */
void calculate_adi_source_dynamics_1d(int nx,double dx2,
                                      double c,double omega,
                                      double *phi,
                                      double *psi,double *L_psi,
                                      double *mu_tot,double *L_mu_tot,
                                      double *xi);

/*!  \brief  calculate the source for 2D ADI for dynamics.
     \param  nx  number of division in x axis
     \param  ny  number of division in y axis
     \param  dx2  square of the size of the mesh in x direction
     \param  dy2  square of the size of the mesh in y direction
     \param  c  factor for ADI
     \param  omega  acceleration factor
     \param  phi  source for ADI (density field)
     \param  psi  psi field
     \param  L_psi  Laplacian of \a psi
     \param  mu_tot  chemical potential
     \param  L_mu_tot  Laplacian of \a mu_tot  */
void calculate_adi_source_dynamics_2d_x(int nx,int ny,
                                        double dx2,
                                        double c,double omega,
                                        double *phi,
                                        double *psi,double *L_psi,
                                        double *mu_tot,double *L_mu_tot,
                                        double *xi);
void calculate_adi_source_dynamics_2d_y(int nx,int ny,
                                        double dy2,
                                        double c,double omega,
                                        double *phi,
                                        double *psi,double *L_psi,
                                        double *mu_tot,double *L_mu_tot,
                                        double *xi);

/*!  \brief  calculate the source for 3D ADI for dynamics.
     \param  nx  number of division in x axis
     \param  ny  number of division in y axis
     \param  nz  number of division in z axis
     \param  dx2  square of the size of the mesh in x direction
     \param  dy2  square of the size of the mesh in y direction
     \param  dz2  square of the size of the mesh in z direction
     \param  c  factor for ADI
     \param  omega  acceleration factor
     \param  phi  source for ADI (density field)
     \param  psi  psi field
     \param  L_psi  Laplacian of \a psi
     \param  mu_tot  chemical potential
     \param  L_mu_tot  Laplacian of \a mu_tot  */
void calculate_adi_source_dynamics_3d_x(int nx,int ny,int nz,
                                        double dx2,
                                        double c,double omega,
                                        double *phi,
                                        double *psi,double *L_psi,
                                        double *mu_tot,double *L_mu_tot,
                                        double *xi);
void calculate_adi_source_dynamics_3d_y(int nx,int ny,int nz,
                                        double dy2,
                                        double c,double omega,
                                        double *phi,
                                        double *psi,double *L_psi,
                                        double *mu_tot,double *L_mu_tot,
                                        double *xi);
void calculate_adi_source_dynamics_3d_z(int nx,int ny,int nz,
                                        double dz2,
                                        double c,double omega,
                                        double *phi,
                                        double *psi,double *L_psi,
                                        double *mu_tot,double *L_mu_tot,
                                        double *xi);

/*---------------------------------------------------------------------------*/
#endif

/*---------------------------------------------------------------------------*/
