/*---------------------------------------------------------------------------*/

/*!  \file  chemicalpotential.c
     \brief  calculate the chemical potential.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005-2007 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "xmalloc.h"
#include "chemicalpotential.h"
#include "ln.h"
#include "nrand.h"
#include "matrix.h"

/*---------------------------------------------------------------------------*/
static void calculate_chemical_potential_local(polymer_blend *blend,
                                                  density_field *density,
                                                  chemical_potential *potential,
                                                  int n);
static void calculate_chemical_potential_longrange_gradient(polymer_blend *blend,
                                                    density_field *density,
                                                    chemical_potential *potential,
                                                    int n);
static void calculate_chemical_potential_longrange(polymer_blend *blend,
                                                    density_field *density,
                                                    chemical_potential *potential,
                                                    int n);

/*---------------------------------------------------------------------------*/
chemical_potential *allocate_chemical_potential(polymer_blend *blend,
                                                field_geometry *geometry)
{
    int np,*ns,nc;
    int n;
    int p,pi;
    chemical_potential *potential;

    np = blend->np;
    ns = blend->ns;
    nc = block_number(blend);
    n = lattice_point_number(geometry);

    potential = xmalloc(sizeof(chemical_potential));

    potential->blend = blend;
    potential->geometry = geometry;

    potential->kappa = xmalloc(sizeof(double) * n);
    potential->mu_kappa = xmalloc(sizeof(double) * n);

    potential->mu = xmalloc(sizeof(double **) * np);
    potential->L_mu_tot = xmalloc(sizeof(double **) * np);
    potential->mu_ext = xmalloc(sizeof(double **) * np);

    potential->mu[0] = xmalloc(sizeof(double *) * nc);
    potential->L_mu_tot[0] = xmalloc(sizeof(double *) * nc);
    potential->mu_ext[0] = xmalloc(sizeof(double *) * nc);

    for(p = 1;p < np;p++)
    {
        potential->mu[p] = potential->mu[p - 1] + ns[p - 1];
        potential->L_mu_tot[p] = potential->L_mu_tot[p - 1] + ns[p - 1];
        potential->mu_ext[p] = potential->mu_ext[p - 1] + ns[p - 1];
    }

    for(pi = 0;pi < nc;pi++)
    {
        (*potential->mu)[pi] = allocate_vector(n);
        (*potential->L_mu_tot)[pi] = allocate_vector(n);
        (*potential->mu_ext)[pi] = allocate_vector(n);
    }

        /*  mu_tot and mu_ext share the same memory space.  */
    potential->mu_tot = potential->mu_ext;
        /*  work_kappa and mu_kappa share the same memory space.  */
    potential->work_kappa = potential->mu_kappa;

    return potential;
}

/*---------------------------------------------------------------------------*/
void free_chemical_potential(chemical_potential *potential)
{
    int nc;
    int pi;

    if(potential == NULL)
        return;

    nc = block_number(potential->blend);

    for(pi = 0;pi < nc;pi++)
    {
        free_vector((*potential->mu)[pi]);
        free_vector((*potential->L_mu_tot)[pi]);
        free_vector((*potential->mu_ext)[pi]);
    }

    free(potential->mu[0]);
    free(potential->L_mu_tot[0]);
    free(potential->mu_ext[0]);

    free(potential->kappa);
    free(potential->mu_kappa);
    free(potential->mu);
    free(potential->L_mu_tot);
    free(potential->mu_ext);

    free(potential);
}

/*---------------------------------------------------------------------------*/
void calculate_chemical_potential(polymer_blend *blend,
                                  density_field *density,
                                  chemical_potential *potential,
                                  field_fftw_plan *plan)
{
    int i,p;
    int n;
    int np,*ns;

    np = blend->np;
    ns = blend->ns;
    n = lattice_point_number(potential->geometry);


        /*  calculate the Green function and the Laplacian of the field.  */
    for(p = 0;p < np;p++)
    {
        if(ns[p] == 1)
        {
            operate_L(potential->geometry,
                      density->psi[p][0],density->L_psi[p][0]);
        }
        else
        {
            for(i = 0;i < ns[p];i++)
            {
                if(blend->chain[p]->lambda > 0.0)
                {
                    operate_Li_cut(potential->geometry,
                                   density->psi[p][i],density->Li_psi[p][i],
                                   blend->chain[p]->lambda,plan);
                }
                else
                {
                    operate_Li(potential->geometry,
                               density->psi[p][i],density->Li_psi[p][i],plan);
                }
                operate_L(potential->geometry,
                          density->psi[p][i],density->L_psi[p][i]);
            }
        }
    }

        /*  calculate the chemical potential.  */
    calculate_chemical_potential_local(blend,density,potential,n);
    calculate_chemical_potential_longrange_gradient(blend,density,potential,n);
}

/*---------------------------------------------------------------------------*/
void calculate_chemical_potential_except_gradient(polymer_blend *blend,
                                  density_field *density,
                                  chemical_potential *potential,
                                  field_fftw_plan *plan)
{
    int i,p;
    int n;
    int np,*ns;

    np = blend->np;
    ns = blend->ns;
    n = lattice_point_number(potential->geometry);


        /*  calculate the Green function and the Laplacian of the field.  */
    for(p = 0;p < np;p++)
    {
        if(ns[p] > 1)
        {
            for(i = 0;i < ns[p];i++)
            {
                if(blend->chain[p]->lambda > 0.0)
                {
/*                     operate_Li_cut(potential->geometry, */
/*                                    density->psi[p][i],density->Li_psi[p][i], */
/*                                    blend->chain[p]->lambda,plan); */
                    operate_Li_cut_for_statics(potential->geometry,
                                               density->psi[p][i],density->Li_psi[p][i],
                                               blend->chain[p]->lambda,plan);
                }
                else
                {
                    operate_Li(potential->geometry,
                               density->psi[p][i],density->Li_psi[p][i],plan);
                }
            }
        }
    }

        /*  calculate the chemical potential.  */
    calculate_chemical_potential_local(blend,density,potential,n);
    calculate_chemical_potential_longrange(blend,density,potential,n);
}

/*---------------------------------------------------------------------------*/
void calculate_chemical_potential_except_gradient_multigrid(polymer_blend *blend,
                                                            density_field *density,
                                                            chemical_potential *potential,
                                                            multigrid_condition *mgcondition,
                                                            multigrid_field *multigrid)
{
    int i,p;
    int n;
    int np,*ns;

    np = blend->np;
    ns = blend->ns;
    n = lattice_point_number(potential->geometry);


        /*  calculate the Green function and the Laplacian of the field.  */
    for(p = 0;p < np;p++)
    {
        if(ns[p] > 1)
        {
            for(i = 0;i < ns[p];i++)
            {
                operate_Li_multigrid(multigrid,mgcondition,
                                     density->psi[p][i],density->Li_psi[p][i]);
            }
        }
    }

        /*  calculate the chemical potential.  */
    calculate_chemical_potential_local(blend,density,potential,n);
    calculate_chemical_potential_longrange(blend,density,potential,n);
}

/*---------------------------------------------------------------------------*/
void calculate_chemical_potential_incompressible(polymer_blend *blend,
                                                 density_field *density,
                                                 chemical_potential *potential)
{
    int p,i,r;
    int np,*ns;
    int nc;
    int n;
    double ***psi;
    double *mu_kappa;
    double psi2;

    np = blend->np;
    ns = blend->ns;
    nc = block_number(blend);
    n = lattice_point_number(potential->geometry);
    psi = density->psi;
    mu_kappa = potential->mu_kappa;


        /*  calculate the Lagrange multipliers
            for the incompressible condition.  */
    for(r = 0;r < n;r++)
    {
        psi2 = 0.0;
        for(p = 0;p < np;p++)
        {
            for(i = 0;i < ns[p];i++)
            {
                psi2 += psi[p][i][r] * psi[p][i][r];
            }
        }

        mu_kappa[r] = (psi2 - 1.0) / nc;
    }
}

/*---------------------------------------------------------------------------*/
void calculate_chemical_potential_local(polymer_blend *blend,
                                        density_field *density,
                                        chemical_potential *potential,int n)
{
    int p,q,i,j;
    int r;
    int np,*ns;
    double ***psi;
    double ***mu;
    double **C,****chi;
    double *f;

    np = blend->np;
    ns = blend->ns;
    psi = density->psi;
    mu = potential->mu;
    chi = blend->chi;

    for(p = 0;p < np;p++)
    {
        f = blend->chain[p]->f;
        C = blend->propagator[p]->C;

        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < n;r++)
            {
                    /*  the entropy term.  */
                mu[p][i][r] =
                    + (2.0 * (f[i] * C[i][i])
                       * (2.0 * psi[p][i][r]
                          * LN(psi[p][i][r]) + psi[p][i][r]));

                    /*  cancel the overcount.  */
                mu[p][i][r] -=
                    + (4.0 * (f[i] * C[i][i])
                       * psi[p][i][r]);
            }

            for(j = 0;j < ns[p];j++)
            {
                for(r = 0;r < n;r++)
                {
                        /*  the correction term.  */
                    mu[p][i][r] +=
                        + (4.0 * (sqrt(f[i] * f[j]) * C[i][j])
                           * psi[p][j][r]);
                }
            }

            for(q = 0;q < np;q++)
            {
                for(j = 0;j < ns[q];j++)
                {
                    for(r = 0;r < n;r++)
                    {
                            /*  the chi parameter term.  */
                        mu[p][i][r] +=
                            + (2.0 * chi[p][i][q][j]
                               * psi[p][i][r] * psi[q][j][r] * psi[q][j][r]);
                    }
                }
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void calculate_chemical_potential_longrange_gradient(polymer_blend *blend,
                                      density_field *density,
                                      chemical_potential *potential,
                                      int n)
{
    int p,i,j;
    int r;
    int np,*ns;
    double ***Li_psi,***L_psi;
    double ***mu;
    double **A,**B;
    double *f;

    np = blend->np;
    ns = blend->ns;
    Li_psi = density->Li_psi;
    L_psi = density->L_psi;
    mu = potential->mu;

    for(p = 0;p < np;p++)
    {
        f = blend->chain[p]->f;
        A = blend->propagator[p]->A;
        B = blend->propagator[p]->B;

        for(i = 0;i < ns[p];i++)
        {
            for(j = 0;j < ns[p];j++)
            {
                for(r = 0;r < n;r++)
                {
                        /*  the long range term and the gradient term.  */
                    mu[p][i][r] +=
                        - (4.0 * (sqrt(f[i] * f[j]) * A[i][j])
                           * Li_psi[p][j][r])
                        - (4.0 * (sqrt(f[i] * f[j]) * B[i][j])
                           * L_psi[p][j][r]);
                }
            }
        }
    }
}
/*---------------------------------------------------------------------------*/
void calculate_chemical_potential_longrange(polymer_blend *blend,
                                      density_field *density,
                                      chemical_potential *potential,
                                      int n)
{
    int p,i,j;
    int r;
    int np,*ns;
    double ***Li_psi;
    double ***mu;
    double **A;
    double *f;

    np = blend->np;
    ns = blend->ns;
    Li_psi = density->Li_psi;
    mu = potential->mu;

    for(p = 0;p < np;p++)
    {
        f = blend->chain[p]->f;
        A = blend->propagator[p]->A;

        for(i = 0;i < ns[p];i++)
        {
            for(j = 0;j < ns[p];j++)
            {
                for(r = 0;r < n;r++)
                {
                        /*  the long range term.  */
                    mu[p][i][r] +=
                        - (4.0 * (sqrt(f[i] * f[j]) * A[i][j])
                           * Li_psi[p][j][r]);
                }
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void add_noise_to_chemical_potential(chemical_potential *potential,
                                     double noise)
{
    int p,i,r;
    int np,*ns;
    int n;
    double ***mu_ext;

    np = potential->blend->np;
    ns = potential->blend->ns;
    n = lattice_point_number(potential->geometry);
    mu_ext = potential->mu_ext;

    for(p = 0;p < np;p++)
    {

        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < n;r++)
            {
                mu_ext[p][i][r] += noise * nrand();
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void calculate_chemical_potential_external(density_field *density,
                                           chemical_potential *potential,
                                           double omega)
{
    int p,i,r;
    int np,*ns;
    int n;
    double ***psi;
    double ***mu_ext;
    double *kappa;

    np = potential->blend->np;
    ns = potential->blend->ns;
    n = lattice_point_number(potential->geometry);
    psi = density->psi;
    mu_ext = potential->mu_ext;

    kappa = potential->kappa;

    for(p = 0;p < np;p++)
    {

        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < n;r++)
            {
                mu_ext[p][i][r] = (1.0 / (-omega)) * kappa[r] * psi[p][i][r];
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void calculate_chemical_potential_external_incompressible(density_field *density,
                                                          chemical_potential *potential)
{
    int p,i,r;
    int np,*ns;
    int n;
    double ***psi;
    double ***mu_ext;
    double *kappa;

    np = potential->blend->np;
    ns = potential->blend->ns;
    n = lattice_point_number(potential->geometry);
    psi = density->psi;
    mu_ext = potential->mu_ext;

    kappa = potential->kappa;

    for(p = 0;p < np;p++)
    {

        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < n;r++)
            {
                mu_ext[p][i][r] = kappa[r] * psi[p][i][r];
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void clear_chemical_potential(chemical_potential *potential)
{
    int r;
    int n;

    n = lattice_point_number(potential->geometry);

    for(r = 0;r < n;r++)
    {
        potential->kappa[r] = 0.0;
    }
}

/*---------------------------------------------------------------------------*/
void estimate_multiplier(polymer_blend *blend,
                         density_field *density,
                         chemical_potential *potential)
{
    int p,i,r;
    int np,*ns;
    int n;
    double ***psi;
    double ***mu;
    double *kappa;
    double psi_tot;
    double mu_tot;

    np = blend->np;
    ns = blend->ns;
    n = lattice_point_number(potential->geometry);
    psi = density->psi;
    mu = potential->mu;
    kappa = potential->kappa;


    for(r = 0;r < n;r++)
    {
        psi_tot = 0.0;
        mu_tot = 0.0;
        for(p = 0;p < np;p++)
        {
            for(i = 0;i < ns[p];i++)
            {
                psi_tot += psi[p][i][r];
                mu_tot += mu[p][i][r];
            }
        }
        kappa[r] = mu_tot / psi_tot;
    }
}

/*---------------------------------------------------------------------------*/
void calculate_mu_tot(chemical_potential *potential)
{
    int p,i,r;
    int np,*ns,n;

    np = potential->blend->np;
    ns = potential->blend->ns;
    n = lattice_point_number(potential->geometry);

    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < n;r++)
            {
                potential->mu_tot[p][i][r] += potential->mu[p][i][r];
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void calculate_L_mu_tot(chemical_potential *potential)
{
    int p,i;
    int np,*ns;

    np = potential->blend->np;
    ns = potential->blend->ns;

    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            operate_L(potential->geometry,
                      potential->mu_tot[p][i],potential->L_mu_tot[p][i]);
        }
    }
}

/*---------------------------------------------------------------------------*/
void calculate_multiplier_incompressible(density_field *density,
                                         chemical_potential *potential,
                                         field_fftw_plan *plan,
                                         double omega)
{
    int p,i,r;
    int np,*ns;
    int n;
    double ***phi;
    double *kappa;
    double *work_kappa;

    np = potential->blend->np;
    ns = potential->blend->ns;
    n = lattice_point_number(potential->geometry);
    phi = density->phi;

    kappa = potential->kappa;
    work_kappa = potential->work_kappa;

    for(r = 0;r < n;r++)
    {
        work_kappa[r] = 1.0;
        for(p = 0;p < np;p++)
        {
            for(i = 0;i < ns[p];i++)
            {
                work_kappa[r] -= phi[p][i][r];
            }
        }
        work_kappa[r] /= omega;
    }

    operate_Li(potential->geometry,
               work_kappa,kappa,plan);
}

/*---------------------------------------------------------------------------*/
