/*---------------------------------------------------------------------------*/

/*!  \file  chemicalpotential.h
     \brief  calculate the chemical potential.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005,2006 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#ifndef INCLUDE_CHEMICALPOTENTIAL_H
#define INCLUDE_CHEMICALPOTENTIAL_H

#ifdef HAVE_FFTW3_H
#include <fftw3.h>
#endif

#include "operator.h"
#include "density.h"
#include "mgpoisson.h"

/*---------------------------------------------------------------------------*/
/*!  \brief  chemical potential field.  */
typedef struct chemical_potential
{
        /*!  \brief  polymer blend.  */
    polymer_blend *blend;
        /*!  \brief  geometry for the field.  */
    field_geometry *geometry;

        /*!  \brief  Lagrange multiplier for incompressible condition.  */
    double *kappa,*mu_kappa;
        /*!  \brief  work array for calculating \a kappa [\mu_kappa].  */
    double *work_kappa;

        /*!  \brief  chemical potential field.  */
    double ***mu;
        /*!  \brief  chemical potential field by external condition.  */
    double ***mu_ext;

        /*!  \brief  total chemical potential (\a mu + \a mu_ext) [\mu_ext].  */
    double ***mu_tot;
        /*!  \brief  Laplacian of \a mu_tot.  */
    double ***L_mu_tot;
} chemical_potential;

/*---------------------------------------------------------------------------*/
/*!  \brief  allocate a new chemical potential.
     \param  blend  polymer blend
     \param  geometry  geometry for the field  */
chemical_potential *allocate_chemical_potential(polymer_blend *blend,
                                                field_geometry *geometry);
/*!  \brief  free a chemical potential.
     \param  potential  chemical potential to free  */
void free_chemical_potential(chemical_potential *potential);

/*!  \brief  calculate chemical potential.
     \param  blend  polymer blend
     \param  density  density field
     \param  potential  chemical potential field
     \param  geometry  geometry for the field
     \param  plan  plan for FFT  */
void calculate_chemical_potential(polymer_blend *blend,
                                  density_field *density,
                                  chemical_potential *potential,
                                  field_fftw_plan *plan);
/*!  \brief  calculate chemical potential except for the gradient term.
     \param  blend  polymer blend
     \param  density  density field
     \param  potential  chemical potential field
     \param  geometry  geometry for the field
     \param  plan  plan for FFT  */
void calculate_chemical_potential_except_gradient(polymer_blend *blend,
                                                  density_field *density,
                                                  chemical_potential *potential,
                                                  field_fftw_plan *plan);
/*!  \brief  calculate chemical potential except for the gradient term.
     \param  blend  polymer blend
     \param  density  density field
     \param  potential  chemical potential field
     \param  geometry  geometry for the field
     \param  mgcondition multigrid Poisson solver condition
     \param  multigrid  multigrid field  */
void calculate_chemical_potential_except_gradient_multigrid(polymer_blend *blend,
                                                            density_field *density,
                                                            chemical_potential *potential,
                                                            multigrid_condition *mgcondition,
                                                            multigrid_field *multigrid);

/*!  \brief  calculate Lagrangian multiplier for incompressible condition.
     \param  blend  polymer blend
     \param  density  density field
     \param  potential  chemical potential field  */
void calculate_chemical_potential_incompressible(polymer_blend *blend,
                                                 density_field *density,
                                                 chemical_potential *potential);

/*!  \brief  add noise to chemical potential.
     \param  potential  chemical potential field
     \param  noise  magnification of noise  */
void add_noise_to_chemical_potential(chemical_potential *potential,
                                     double noise);

/*!  \brief  calculate chemical potential from external constraints.
     \param  potential  chemical potential field  */
void calculate_chemical_potential_external(density_field *density,
                                           chemical_potential *potential,
                                           double omega);
/*!  \brief  calculate chemical potential from external incompressible constraint.
     \param  potential  chemical potential field  */
void calculate_chemical_potential_external_incompressible(density_field *density,
                                                          chemical_potential *potential);

/*!  \brief  clear chemical potential (set it to 0).
     \param  potential  chemical potential field  */
void clear_chemical_potential(chemical_potential *potential);

/*!  \brief  estimate Lagrangian multiplier.
     \param  blend  polymer blend
     \param  density  density field
     \param  potential  chemical potential field  */
void estimate_multiplier(polymer_blend *blend,
                         density_field *density,
                         chemical_potential *potential);

/*!  \brief  calculate total chemical potential \a mu_tot. (\a mu_ext is overwritten).
     \param  potential  chemical potential field  */
void calculate_mu_tot(chemical_potential *potential);
/*!  \brief  calculate Laplacian of chemical potential.
     \param  potential  chemical potential field  */
void calculate_L_mu_tot(chemical_potential *potential);

/*!  \brief  calculate Lagrangian multiplier for incompressible condition (\a kappa).
     \param  density  density field
     \param  potential  chemical potential field
     \param  plan  plan for FFT
     \param  omega  accelaration factor  */
void calculate_multiplier_incompressible(density_field *density,
                                         chemical_potential *potential,
                                         field_fftw_plan *plan,
                                         double omega);

/*---------------------------------------------------------------------------*/
#endif

/*---------------------------------------------------------------------------*/
