/*---------------------------------------------------------------------------*/

/*!  \file  density.c
     \brief  the library for the polymer density field.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005,2006,2014 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include "xmalloc.h"
#include "density.h"
#include "mt.h"
#include "nrand.h"
#include "matrix.h"

/*---------------------------------------------------------------------------*/
density_field *allocate_density_field(polymer_blend *blend,
                                      field_geometry *geometry)
{
    int np,*ns,nc;
    int p,pi;
    int n;
    density_field *density;

    np = blend->np;
    ns = blend->ns;
    nc = block_number(blend);
    n = lattice_point_number(geometry);

    density = xmalloc(sizeof(density_field));

    density->blend = blend;
    density->geometry = geometry;

    density->work1 = xmalloc(sizeof(double **) * np);
    density->work2 = xmalloc(sizeof(double **) * np);
    density->L_psi = xmalloc(sizeof(double **) * np);
    density->Li_psi = xmalloc(sizeof(double **) * np);
    density->xi = xmalloc(sizeof(double **) * np);

    density->work1[0] = xmalloc(sizeof(double *) * nc);
    density->work2[0] = xmalloc(sizeof(double *) * nc);
    density->L_psi[0] = xmalloc(sizeof(double *) * nc);
    density->Li_psi[0] = xmalloc(sizeof(double *) * nc);
    density->xi[0] = xmalloc(sizeof(double *) * nc);
    for(p = 1;p < np;p++)
    {
        density->work1[p] = density->work1[p - 1] + ns[p - 1];
        density->work2[p] = density->work2[p - 1] + ns[p - 1];
        density->L_psi[p] = density->L_psi[p - 1] + ns[p - 1];
        density->Li_psi[p] = density->Li_psi[p - 1] + ns[p - 1];
        density->xi[p] = density->xi[p - 1] + ns[p - 1];
    }
    for(pi = 0;pi < nc;pi++)
    {
        (*density->work1)[pi] = allocate_vector(n);
        (*density->work2)[pi] = allocate_vector(n);
        (*density->L_psi)[pi] = allocate_vector(n);
        (*density->Li_psi)[pi] = allocate_vector(n);
        (*density->xi)[pi] = allocate_vector(n);
    }

        /*  psi and work1 (or work2) share the same memory space.
            work1 and work2 will be swapped for the ADI scheme.  */
    density->psi = density->work1;
        /*  phi and work2 (or work1) share the same memory space.  */
    density->phi = get_unused_work_array(density->psi,density);

    return density;
}

/*---------------------------------------------------------------------------*/
void free_density_field(density_field *density)
{
    int nc;
    int pi;

    if(density == NULL)
        return;

    nc = block_number(density->blend);

    for(pi = 0;pi < nc;pi++)
    {
        free_vector((*density->work1)[pi]);
        free_vector((*density->work2)[pi]);
        free_vector((*density->L_psi)[pi]);
        free_vector((*density->Li_psi)[pi]);
        free_vector((*density->xi)[pi]);
    }
    free(density->work1[0]);
    free(density->work2[0]);
    free(density->L_psi[0]);
    free(density->Li_psi[0]);
    free(density->xi[0]);

    free(density->work1);
    free(density->work2);
    free(density->L_psi);
    free(density->Li_psi);
    free(density->xi);

    free(density);
}

/*---------------------------------------------------------------------------*/
void set_density_field_random(density_field *density,double noise)
{
    int p,i,r;
    int np,*ns;
    int n;
    double psi_bar;

    ns = density->blend->ns;
    np = density->blend->np;
    n = lattice_point_number(density->geometry);


        /*  white noise.  */
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            psi_bar = psi_spatial_average(density->blend,p,i);
            for(r = 0;r < n;r++)
            {
                density->psi[p][i][r] = fabs(psi_bar + noise * nrand());
            }
        }
    }

    reset_density_field_offset(density);
}

/*---------------------------------------------------------------------------*/
void reset_density_field_offset(density_field *density)
{
    int p,i,r;
    int np,*ns;
    int n;
    double ***psi;
    double psi2,psi_bar2,psi_bar;
    double lambda;

    ns = density->blend->ns;
    np = density->blend->np;
    n = lattice_point_number(density->geometry);
    psi = density->psi;

    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            psi2 = 0.0;
            psi_bar2 = psi_spatial_average(density->blend,p,i) *
                psi_spatial_average(density->blend,p,i) * n;
            for(r = 0;r < n;r++)
            {
                psi2 += psi[p][i][r] * psi[p][i][r];
            }

            if(psi2 != 0.0)
            {
                lambda = sqrt(psi_bar2 / psi2);

                for(r = 0;r < n;r++)
                {
                    psi[p][i][r] *= lambda;
                }
            }
            else
            {
                psi_bar = psi_spatial_average(density->blend,p,i);
                for(r = 0;r < n;r++)
                {
                    psi[p][i][r] = psi_bar;
                }
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void correct_density_field(density_field *density)
{
    int np,*ns;
    int n;
    int p,i,r;
    double ***psi;

    np = density->blend->np;
    ns = density->blend->ns;
    n = lattice_point_number(density->geometry);

    psi = density->psi;

    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < n;r++)
            {
                    psi[p][i][r] = fabs(psi[p][i][r]);
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void normalize_density_field(density_field *density)
{
    int np,*ns,n;
    int p,i,r;
    double ***psi;
    double phi,phi_bar;

    np = density->blend->np;
    ns = density->blend->ns;
    n = lattice_point_number(density->geometry);

    psi = density->psi;

    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            phi = 0;
            for(r = 0;r < n;r++)
            {
                phi += psi[p][i][r] * psi[p][i][r];
            }
            phi /= n;

            phi_bar = density->blend->volume_fraction[p] * density->blend->chain[p]->f[i];

            for(r = 0;r < n;r++)
            {
                psi[p][i][r] *= sqrt(phi_bar / phi);
            }
        }
    }

    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < n;r++)
            {
                if(psi[p][i][r] > 1)
                    psi[p][i][r] = 1;
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
double ***get_unused_work_array(double ***psi,density_field *density)
{
    if(psi == density->work1)
        return density->work2;
    else
        return density->work1;
}

/*---------------------------------------------------------------------------*/
void calculate_phi_from_psi(density_field *density)
{
    int np,*ns,n;
    int p,i,r;
    double ***psi;
    double ***phi;

    np = density->blend->np;
    ns = density->blend->ns;
    n = lattice_point_number(density->geometry);

    density->phi = get_unused_work_array(density->psi,density);
    psi = density->psi;
    phi = density->phi;

    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < n;r++)
            {
                phi[p][i][r] = psi[p][i][r] * psi[p][i][r];
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void calculate_psi_from_phi(density_field *density)
{
    int np,*ns,n;
    int p,i,r;
    double ***psi;
    double ***phi;

    np = density->blend->np;
    ns = density->blend->ns;
    n = lattice_point_number(density->geometry);

    psi = density->psi;
    phi = density->phi;

    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < n;r++)
            {
                if(phi[p][i][r] >= 0.0)
                {
                    if(phi[p][i][r] <= 1.0)
                    {
                        psi[p][i][r] = sqrt(phi[p][i][r]);
                    }
                    else
                    {
                        psi[p][i][r] = 1.0;
                    }
                }
                else
                {
                        /*  if the value of the density field is negative,
                            the psi-field value is assumed to be zero.  */
                    psi[p][i][r] = 0.0;
                }
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
