/*---------------------------------------------------------------------------*/

/*!  \file  density.h
     \brief  the library for the polymer density field.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005,2006 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#ifndef INCLUDE_DENSITY_H
#define INCLUDE_DENSITY_H

#ifdef HAVE_FFTW3_H
#include <fftw3.h>
#endif

#include "geometry.h"
#include "blend.h"

/*---------------------------------------------------------------------------*/
/*!  \brief  segment density field.  */
typedef struct density_field
{
        /*!  \brief  polymer blend.  */
    polymer_blend *blend;
        /*!  \brief  geometry for the field.  */
    field_geometry *geometry;

        /*!  \brief  work arrays.  */
    double ***work1;
    double ***work2;

        /*!  \brief  psi field (square root of \a phi) [\a work1 or \a work2].  */
    double ***psi;
        /*!  \brief  Laplacian of \a psi.  */
    double ***L_psi;
        /*!  \brief  Laplacian inverse of \a psi.  */
    double ***Li_psi;

        /*!  \brief  segment density field  [\a work1 or \a work2].  */
    double ***phi;

        /*!  \brief  random force field.  */
    double ***xi;
} density_field;

/*---------------------------------------------------------------------------*/
/*!  \biref  allocate a new density field.
     \param  blend  polymer blend
     \param  geometry  geometry of the field  */
density_field *allocate_density_field(polymer_blend *blend,
                                      field_geometry *geometry);
/*!  \biref  free a density field.
     \param  density  density field to free  */
void free_density_field(density_field *density);

/*!  \biref  set density field randomly.
     \param  density  density field
     \param  noise  magnification of noise  */
void set_density_field_random(density_field *density,double noise);
/*!  \biref  reset density field to correct its mean value.
     \param  density  density field
     \param  blend  polymer blend  */
void reset_density_field_offset(density_field *density);

/*!  \biref  correct \a phi field to be positive.
     \param  density  density fieldd  */
void correct_density_field(density_field *density);

/*!  \biref  rescales density field for mass conservation.
     \param  density  density field  */
void normalize_density_field(density_field *density);
/*!  \biref  get the unused work array for density or psi field.
     \param  psi  psi field
     \param  density  density field  */
double ***get_unused_work_array(double ***psi,density_field *density);

/*!  \brief  calculate the \a phi field from the \a psi field.
     \param  density   density field  */
void calculate_phi_from_psi(density_field *density);
/*!  \brief  calculate the \a psi field from the \a phi field.
     \param  density   density field  */
void calculate_psi_from_phi(density_field *density);

/*---------------------------------------------------------------------------*/
#endif

/*---------------------------------------------------------------------------*/
