/*---------------------------------------------------------------------------*/

/*!  \file  diffuser.c
     \brief  a field evolver for dynamics, based on the diffusion.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005-2007 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#ifdef HAVE_FFTW3_H
#include <fftw3.h>
#endif

#include "operator.h"
#include "tridiag.h"
#include "diffuser.h"
#include "adi.h"
#include "noise.h"

/*---------------------------------------------------------------------------*/
static void diffuse_density_adi_1d(density_field *density,
                                   chemical_potential *potential,
                                   external_force *force,
                                   double omega,double noise,
                                   field_fftw_plan *plan,multigrid_condition *mgcondition,
                                   multigrid_field *multigrid,int use_multigrid,
                                   tridiag_matrix *tridiag,double *u);
static void diffuse_density_adi_2d(density_field *density,
                                   chemical_potential *potential,
                                   external_force *force,
                                   double omega,double noise,
                                   field_fftw_plan *plan,multigrid_condition *mgcondition,
                                   multigrid_field *multigrid,int use_multigrid,
                                   tridiag_matrix *tridiag,double *u);
static void diffuse_density_adi_3d(density_field *density,
                                   chemical_potential *potential,
                                   external_force *force,
                                   double omega,double noise,
                                   field_fftw_plan *plan,multigrid_condition *mgcondition,
                                   multigrid_field *multigrid,int use_multigrid,
                                   tridiag_matrix *tridiag,double *u);
static void set_tridiagonal_matrix(int n,double h,tridiag_matrix *tridiag);
static void calculate_chemical_potential_for_adi(density_field *density,
                                                 chemical_potential *potential,
                                                 external_force *force,
                                                 double omega,double noise,
                                                 field_fftw_plan *plan,
                                                 multigrid_condition *mgcondition,
                                                 multigrid_field *multigrid,
                                                 int use_multigrid);

/*---------------------------------------------------------------------------*/
void diffuse_density(density_field *density,
                     chemical_potential *potential,
                     double omega)
{
    int p,i;
    int n;
    int r;
    int np,*ns;
    double ***phi;
    double ***psi,***L_psi;
    double ***mu_tot,***L_mu_tot;
    double ***xi;

    np = density->blend->np;
    ns = density->blend->ns;
    n = lattice_point_number(density->geometry);

    psi = density->psi;
    L_psi = density->L_psi;
    phi = density->phi;
    mu_tot = potential->mu_tot;
    L_mu_tot = potential->L_mu_tot;
    xi = density->xi;

    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < n;r++)
            {
                phi[p][i][r] = psi[p][i][r] * psi[p][i][r]
                    + omega * (+ psi[p][i][r] * L_mu_tot[p][i][r]
                               - L_psi[p][i][r] * mu_tot[p][i][r])
                    + xi[p][i][r];
            }
        }
    }

    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < n;r++)
            {
                if(phi[p][i][r] > 0)
                {
                    psi[p][i][r] = sqrt(phi[p][i][r]);
                }
                else
                {
                    psi[p][i][r] = 0;
                }
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void diffuse_density_adi(density_field *density,
                         chemical_potential *potential,
                         external_force *force,
                         double omega,double noise,
                         field_fftw_plan *plan,multigrid_condition *mgcondition,
                         multigrid_field *multigrid,int use_multigrid,
                         tridiag_matrix *tridiag,double *u)
{
    switch(density->geometry->dimension)
    {
        case 1:
            diffuse_density_adi_1d(density,potential,force,omega,noise,plan,
                                   mgcondition,multigrid,use_multigrid,tridiag,u);
            break;

        case 2:
            diffuse_density_adi_2d(density,potential,force,omega,noise,plan,
                                   mgcondition,multigrid,use_multigrid,tridiag,u);
            break;

        case 3:
            diffuse_density_adi_3d(density,potential,force,omega,noise,plan,
                                   mgcondition,multigrid,use_multigrid,tridiag,u);
            break;

        default:
            return;
    }
}

/*---------------------------------------------------------------------------*/
void diffuse_density_adi_1d(density_field *density,
                            chemical_potential *potential,
                            external_force *force,
                            double omega,double noise,
                            field_fftw_plan *plan,multigrid_condition *mgcondition,
                            multigrid_field *multigrid,int use_multigrid,
                            tridiag_matrix *tridiag,double *u)
{
    int p,i;
    int nx;
    int np,*ns;
    double ***psi;
    double ***phi;
    double c;
    double dx;
    double ***mu_tot,***L_mu_tot;
    double ***L_psi;
    double ***xi;

    np = density->blend->np;
    ns = density->blend->ns;
    nx = density->geometry->nx;
    dx = density->geometry->lx / density->geometry->nx;

    psi = density->psi;
    phi = density->phi;
    L_psi = density->L_psi;
    mu_tot = potential->mu_tot;
    L_mu_tot = potential->L_mu_tot;
    xi = density->xi;

    calculate_chemical_potential_for_adi(density,
                                         potential,
                                         force,
                                         omega,noise,
                                         plan,
                                         mgcondition,
                                         multigrid,
                                         use_multigrid);


        /*  evolve density field.  */
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            c = omega *
                (2 * density->blend->chain[p]->f[i]
                 * density->blend->propagator[p]->C[i][i]);

                /*  calculate source term for the ADI method.  */
            calculate_adi_source_dynamics_1d(nx,dx * dx,
                                             c,omega,
                                             phi[p][i],
                                             psi[p][i],L_psi[p][i],
                                             mu_tot[p][i],L_mu_tot[p][i],
                                             xi[p][i]);

                /*  set tridiagonal matrix.  */
            set_tridiagonal_matrix(nx,c / (dx * dx),tridiag);

            solve_cyclic_tridiag(nx,1,1,
                                 tridiag,
                                 psi[p][i],u);
        }
    }

    calculate_psi_from_phi(density);
}

/*---------------------------------------------------------------------------*/
void diffuse_density_adi_2d(density_field *density,
                            chemical_potential *potential,
                            external_force *force,
                            double omega,double noise,
                            field_fftw_plan *plan,multigrid_condition *mgcondition,
                            multigrid_field *multigrid,int use_multigrid,
                            tridiag_matrix *tridiag,double *u)
{
    int p,i;
    int nx,ny;
    int np,*ns;
    double ***psi;
    double ***phi;
    double c;
    double dx,dy;
    double ***mu_tot,***L_mu_tot;
    double ***L_psi;
    double ***xi;

    np = density->blend->np;
    ns = density->blend->ns;

    nx = density->geometry->nx;
    ny = density->geometry->ny;
    dx = density->geometry->lx / density->geometry->nx;
    dy = density->geometry->ly / density->geometry->ny;

    psi = density->psi;
    phi = density->phi;
    L_psi = density->L_psi;
    mu_tot = potential->mu_tot;
    L_mu_tot = potential->L_mu_tot;
    xi = density->xi;

    omega /= 2;


    calculate_chemical_potential_for_adi(density,
                                         potential,
                                         force,
                                         omega,noise,
                                         plan,
                                         mgcondition,
                                         multigrid,
                                         use_multigrid);

        /*  evolve density field  for x axis.  */
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            c = omega *
                (2 * density->blend->chain[p]->f[i]
                 * density->blend->propagator[p]->C[i][i]);

            calculate_adi_source_dynamics_2d_x(nx,ny,dx * dx,
                                               c,omega,
                                               phi[p][i],
                                               psi[p][i],L_psi[p][i],
                                               mu_tot[p][i],L_mu_tot[p][i],
                                               xi[p][i]);

            set_tridiagonal_matrix(nx,c / (dx * dx),tridiag);

            solve_cyclic_tridiag(nx,ny,ny,
                                 tridiag,
                                 phi[p][i],u);
        }
    }
    calculate_psi_from_phi(density);


    calculate_chemical_potential_for_adi(density,
                                         potential,
                                         force,
                                         omega,noise,
                                         plan,
                                         mgcondition,
                                         multigrid,
                                         use_multigrid);


        /*  evolve density field for y axis.  */
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            c = omega *
                (2 * density->blend->chain[p]->f[i]
                 * density->blend->propagator[p]->C[i][i]);

            calculate_adi_source_dynamics_2d_y(nx,ny,dy * dy,
                                               c,omega,
                                               phi[p][i],
                                               psi[p][i],L_psi[p][i],
                                               mu_tot[p][i],L_mu_tot[p][i],
                                               xi[p][i]);

            set_tridiagonal_matrix(ny,c / (dy * dy),tridiag);

            solve_cyclic_tridiag(ny,nx,1,
                                 tridiag,
                                 phi[p][i],u);
        }
    }
    calculate_psi_from_phi(density);
}

/*---------------------------------------------------------------------------*/
void diffuse_density_adi_3d(density_field *density,
                            chemical_potential *potential,
                            external_force *force,
                            double omega,double noise,
                            field_fftw_plan *plan,multigrid_condition *mgcondition,
                            multigrid_field *multigrid,int use_multigrid,
                            tridiag_matrix *tridiag,double *u)
{
    int p,i;
    int nx,ny,nz;
    int np,*ns;
    double ***psi;
    double ***phi;
    double c;
    double dx,dy,dz;
    double ***mu_tot,***L_mu_tot;
    double ***L_psi;
    double ***xi;

    np = density->blend->np;
    ns = density->blend->ns;

    nx = density->geometry->nx;
    ny = density->geometry->ny;
    nz = density->geometry->nz;
    dx = density->geometry->lx / density->geometry->nx;
    dy = density->geometry->ly / density->geometry->ny;
    dz = density->geometry->lz / density->geometry->nz;

    psi = density->psi;
    phi = density->phi;
    L_psi = density->L_psi;
    mu_tot = potential->mu_tot;
    L_mu_tot = potential->L_mu_tot;
    xi = density->xi;

    omega /= 3;





    calculate_chemical_potential_for_adi(density,
                                         potential,
                                         force,
                                         omega,noise,
                                         plan,
                                         mgcondition,
                                         multigrid,
                                         use_multigrid);

        /*  evolve density field  for x axis.  */
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            c = omega *
                (2 * density->blend->chain[p]->f[i]
                 * density->blend->propagator[p]->C[i][i]);

            calculate_adi_source_dynamics_3d_x(nx,ny,nz,dx * dx,
                                               c,omega,
                                               phi[p][i],
                                               psi[p][i],L_psi[p][i],
                                               mu_tot[p][i],L_mu_tot[p][i],
                                               xi[p][i]);

            set_tridiagonal_matrix(nx,c / (dx * dx),tridiag);

            solve_cyclic_tridiag(nx,ny * nz,ny * nz,
                                 tridiag,
                                 phi[p][i],u);
        }
    }
    calculate_psi_from_phi(density);


    calculate_chemical_potential_for_adi(density,
                                         potential,
                                         force,
                                         omega,noise,
                                         plan,
                                         mgcondition,
                                         multigrid,
                                         use_multigrid);

        /*  evolve density field  for y axis.  */
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            c = omega *
                (2 * density->blend->chain[p]->f[i]
                 * density->blend->propagator[p]->C[i][i]);

            calculate_adi_source_dynamics_3d_y(nx,ny,nz,dy * dy,
                                               c,omega,
                                               phi[p][i],
                                               psi[p][i],L_psi[p][i],
                                               mu_tot[p][i],L_mu_tot[p][i],
                                               xi[p][i]);

            set_tridiagonal_matrix(ny,c / (dy * dy),tridiag);

            solve_cyclic_tridiag(ny,nx * nz,nz,
                                 tridiag,
                                 phi[p][i],u);
        }
    }
    calculate_psi_from_phi(density);


    calculate_chemical_potential_for_adi(density,
                                         potential,
                                         force,
                                         omega,noise,
                                         plan,
                                         mgcondition,
                                         multigrid,
                                         use_multigrid);

        /*  evolve density field for z axis.  */
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            c = omega *
                (2 * density->blend->chain[p]->f[i]
                 * density->blend->propagator[p]->C[i][i]);

            calculate_adi_source_dynamics_3d_z(nx,ny,nz,dz * dz,
                                               c,omega,
                                               phi[p][i],
                                               psi[p][i],L_psi[p][i],
                                               mu_tot[p][i],L_mu_tot[p][i],
                                               xi[p][i]);

            set_tridiagonal_matrix(nz,c / (dz * dz),tridiag);

            solve_cyclic_tridiag(nz,nx * ny,1,
                                 tridiag,
                                 phi[p][i],u);
        }
    }
    calculate_psi_from_phi(density);
}

/*---------------------------------------------------------------------------*/
void set_tridiagonal_matrix(int n,double h,tridiag_matrix *tridiag)
{
    int i;

    for(i = 0;i < n;i++)
    {
        tridiag->dl[i] = -h;
        tridiag->du[i] = -h;
        tridiag->d[i] = 1 + 2 * h;
    }
}

/*---------------------------------------------------------------------------*/
void calculate_chemical_potential_for_adi(density_field *density,
                                          chemical_potential *potential,
                                          external_force *force,
                                          double omega,double noise,
                                          field_fftw_plan *plan,
                                          multigrid_condition *mgcondition,
                                          multigrid_field *multigrid,
                                          int use_multigrid)
{
        /*  calculate chemical potential.  */
    calculate_chemical_potential(density->blend,density,potential,plan);
    calculate_chemical_potential_external_incompressible(density,
                                                         potential);
    add_force_to_chemical_potential(potential,force);
    calculate_mu_tot(potential);
    calculate_L_mu_tot(potential);

        /*  calculate FDT noise.  */
    calculate_field_noise_dynamics(density,omega * noise * 4.0);

        /*  calculate Lagrange multiplier.  */
    calculate_multiplier_incompressible(density,potential,plan,omega);
}

/*---------------------------------------------------------------------------*/
void correct_negative_density_field_with_threshold(density_field *density,
                                                   double phi_min)
{
    int p,i,r;
    int np,*ns;
    int n;
    double ***phi;

        /*  if phi_min i positive, do nothing.  */
    if(phi_min > 0)
        return;

    ns = density->blend->ns;
    np = density->blend->np;
    n = lattice_point_number(density->geometry);
    phi = density->phi;

    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < n;r++)
            {
                if(phi[p][i][r] < phi_min)
                {
                    reset_density_field_offset(density);
                    calculate_phi_from_psi(density);
                    return;
                }
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
