/*---------------------------------------------------------------------------*/

/*!  \file  dxwrite.c
     \brief  output the field data as OpenDX format.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005-2007 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include "xmalloc.h"
#include "dxwrite.h"

/*---------------------------------------------------------------------------*/
static void write_field_dx_text_1d(char *filename,density_field *density,
                                   chemical_potential *potential);
static void write_field_dx_text_2d(char *filename,density_field *density,
                                   chemical_potential *potential);
static void write_field_dx_text_3d(char *filename,density_field *density,
                                   chemical_potential *potential);
static void write_field_dx_binary_1d(char *filename,density_field *density,
                                     chemical_potential *potential);
static void write_field_dx_binary_2d(char *filename,density_field *density,
                                     chemical_potential *potential);
static void write_field_dx_binary_3d(char *filename,density_field *density,
                                     chemical_potential *potential);
static void write_field_dx_fieldlist(FILE *fp,polymer_blend *blend);
static int digits_in_decimal(int n);

/*---------------------------------------------------------------------------*/
void write_field_dx(char *filename,density_field *density,
                    chemical_potential *potential)
{
    if((filename == NULL) || (strlen(filename) == 0))
        return;

    if(sizeof(float) == 4)
    {
            /*  IEEE floating number format (binary)  */
        switch(density->geometry->dimension)
        {
            case 1:
                write_field_dx_binary_1d(filename,density,potential);
                break;

            case 2:
                write_field_dx_binary_2d(filename,density,potential);
                break;

            case 3:
                write_field_dx_binary_3d(filename,density,potential);
                break;

            default:
                return;
        }
    }
    else
    {
            /*  ASCII text (if IEEE floating number is not available)  */

        switch(density->geometry->dimension)
        {
            case 1:
                write_field_dx_text_1d(filename,density,potential);
                break;

            case 2:
                write_field_dx_text_2d(filename,density,potential);
                break;

            case 3:
                write_field_dx_text_3d(filename,density,potential);
                break;

            default:
                return;
        }
    }
}

/*---------------------------------------------------------------------------*/
void write_field_dx_text_1d(char *filename,density_field *density,
                       chemical_potential *potential)
{
    FILE *fp;
    int p,i,r;
    int number;
    int np,*ns;
    int nx;
    double lx;
    double ***psi,***mu;

    np = density->blend->np;
    ns = density->blend->ns;
    nx = density->geometry->nx;
    lx = density->geometry->lx;
    psi = density->psi;
    mu = potential->mu;

    fp = fopen(filename,"w");

    fprintf(fp,"object \"positions\" class gridpositions counts %d\n",nx);
    fprintf(fp,"origin\t0\n");
    fprintf(fp,"delta\t%g\n",lx / nx);
    fprintf(fp,"\n");

    fprintf(fp,"object \"connections\" class gridconnections counts %d\n",nx);
    fprintf(fp,"\n");


    number = 0;
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            fprintf(fp,"object \"data_phi%d\" class array type double items %d data follows\n",
                    number,nx);
            for(r = 0;r < nx;r++)
            {
                fprintf(fp,"%g ",psi[p][i][r] * psi[p][i][r]);
            }
            fprintf(fp,"\n");

            fprintf(fp,"object \"data_mu%d\" class array type double items %d data follows\n",
                    number,nx);
            for(r = 0;r < nx;r++)
            {
                fprintf(fp,"%g ",mu[p][i][r]);
            }
            fprintf(fp,"\n");

            number++;
        }
    }
    fprintf(fp,"\n");

    write_field_dx_fieldlist(fp,potential->blend);

    number = 0;
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            fprintf(fp,"object \"phi%d\" class field\n",number);
            fprintf(fp,"component \"positions\" value \"positions\"\n");
            fprintf(fp,"component \"connections\" value \"connections\"\n");
            fprintf(fp,"component \"data\" value \"data_phi%d\"\n",number);
            fprintf(fp,"\n");

            fprintf(fp,"object \"mu%d\" class field\n",number);
            fprintf(fp,"component \"positions\" value \"positions\"\n");
            fprintf(fp,"component \"connections\" value \"connections\"\n");
            fprintf(fp,"component \"data\" value \"data_mu%d\"\n",number);
            fprintf(fp,"\n");

            number++;
        }
    }

    fprintf(fp,"end\n");

    fclose(fp);
}

/*---------------------------------------------------------------------------*/
void write_field_dx_text_2d(char *filename,density_field *density,
                       chemical_potential *potential)
{
    FILE *fp;
    int p,i,r;
    int number;
    int np,*ns;
    int nx,ny;
    double lx,ly;
    double ***psi,***mu;

    np = density->blend->np;
    ns = density->blend->ns;
    nx = density->geometry->nx;
    ny = density->geometry->ny;
    lx = density->geometry->lx;
    ly = density->geometry->ly;
    psi = density->psi;
    mu = potential->mu;

    fp = fopen(filename,"w");

    fprintf(fp,"object \"positions\" class gridpositions counts %d %d\n",nx,ny);
    fprintf(fp,"origin\t0\t0\n");
    fprintf(fp,"delta\t%g\t0\n",lx / nx);
    fprintf(fp,"delta\t0\t%g\n",ly / ny);
    fprintf(fp,"\n");

    fprintf(fp,"object \"connections\" class gridconnections counts %d %d\n",nx,ny);
    fprintf(fp,"\n");


    number = 0;
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            fprintf(fp,"object \"data_phi%d\" class array type double items %d data follows\n",
                    number,nx * ny);
            for(r = 0;r < nx * ny;r++)
            {
                fprintf(fp,"%g ",psi[p][i][r] * psi[p][i][r]);
            }
            fprintf(fp,"\n");

            fprintf(fp,"object \"data_mu%d\" class array type double items %d data follows\n",
                    number,nx * ny);
            for(r = 0;r < nx * ny;r++)
            {
                fprintf(fp,"%g ",mu[p][i][r]);
            }
            fprintf(fp,"\n");

            number++;
        }
    }
    fprintf(fp,"\n");

    write_field_dx_fieldlist(fp,potential->blend);

    number = 0;
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            fprintf(fp,"object \"phi%d\" class field\n",number);
            fprintf(fp,"component \"positions\" value \"positions\"\n");
            fprintf(fp,"component \"connections\" value \"connections\"\n");
            fprintf(fp,"component \"data\" value \"data_phi%d\"\n",number);
            fprintf(fp,"\n");

            fprintf(fp,"object \"mu%d\" class field\n",number);
            fprintf(fp,"component \"positions\" value \"positions\"\n");
            fprintf(fp,"component \"connections\" value \"connections\"\n");
            fprintf(fp,"component \"data\" value \"data_mu%d\"\n",number);
            fprintf(fp,"\n");

            number++;
        }
    }

    fprintf(fp,"end\n");

    fclose(fp);
}

/*---------------------------------------------------------------------------*/
void write_field_dx_text_3d(char *filename,density_field *density,
                       chemical_potential *potential)
{
    FILE *fp;
    int p,i,r;
    int number;
    int np,*ns;
    int nx,ny,nz;
    double lx,ly,lz;
    double ***psi,***mu;

    np = density->blend->np;
    ns = density->blend->ns;
    nx = density->geometry->nx;
    ny = density->geometry->ny;
    nz = density->geometry->nz;
    lx = density->geometry->lx;
    ly = density->geometry->ly;
    lz = density->geometry->lz;
    psi = density->psi;
    mu = potential->mu;

    fp = fopen(filename,"w");

    fprintf(fp,"object \"positions\" class gridpositions counts %d %d %d\n",nx,ny,nz);
    fprintf(fp,"origin\t0\t0\t0\n");
    fprintf(fp,"delta\t%g\t0\t0\n",lx / nx);
    fprintf(fp,"delta\t0\t%g\t0\n",ly / ny);
    fprintf(fp,"delta\t0\t0\t%g\n",lz / nz);
    fprintf(fp,"\n");

    fprintf(fp,"object \"connections\" class gridconnections counts %d %d %d\n",nx,ny,nz);
    fprintf(fp,"\n");


    number = 0;
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            fprintf(fp,"object \"data_phi%d\" class array type double items %d data follows\n",
                    number,nx * ny * nz);
            for(r = 0;r < nx * ny * nz;r++)
            {
                fprintf(fp,"%g ",psi[p][i][r] * psi[p][i][r]);
            }
            fprintf(fp,"\n");

            fprintf(fp,"object \"data_mu%d\" class array type double items %d data follows\n",
                    number,nx * ny * nz);
            for(r = 0;r < nx * ny * nz;r++)
            {
                fprintf(fp,"%g ",mu[p][i][r]);
            }
            fprintf(fp,"\n");

            number++;
        }
    }
    fprintf(fp,"\n");

    write_field_dx_fieldlist(fp,potential->blend);

    number = 0;
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            fprintf(fp,"object \"phi%d\" class field\n",number);
            fprintf(fp,"component \"positions\" value \"positions\"\n");
            fprintf(fp,"component \"connections\" value \"connections\"\n");
            fprintf(fp,"component \"data\" value \"data_phi%d\"\n",number);
            fprintf(fp,"\n");

            fprintf(fp,"object \"mu%d\" class field\n",number);
            fprintf(fp,"component \"positions\" value \"positions\"\n");
            fprintf(fp,"component \"connections\" value \"connections\"\n");
            fprintf(fp,"component \"data\" value \"data_mu%d\"\n",number);
            fprintf(fp,"\n");

            number++;
        }
    }

    fprintf(fp,"end\n");

    fclose(fp);
}

/*---------------------------------------------------------------------------*/
void write_field_dx_binary_1d(char *filename,density_field *density,
                              chemical_potential *potential)
{
    FILE *fp;
    int p,i,r;
    int number,offset;
    int np,*ns;
    int nx;
    double lx;
    double ***psi,***mu;
    float *work;
    char *endian;

#ifdef WORDS_BIGENDIAN
    endian = "msb";
#else
    endian = "lsb";
#endif

    np = density->blend->np;
    ns = density->blend->ns;
    nx = density->geometry->nx;
    lx = density->geometry->lx;
    psi = density->psi;
    mu = potential->mu;
    work = xmalloc(sizeof(float) * nx);

    fp = fopen(filename,"w");

    fprintf(fp,"object \"positions\" class gridpositions counts %d\n",nx);
    fprintf(fp,"origin\t0\n");
    fprintf(fp,"delta\t%g\n",lx / nx);
    fprintf(fp,"\n");

    fprintf(fp,"object \"connections\" class gridconnections counts %d\n",nx);
    fprintf(fp,"\n");

    number = 0;
    offset = 0;
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            fprintf(fp,"object \"data_phi%d\" class array type float items %d %s ieee data %d\n",
                    number,nx,endian,offset);
            fprintf(fp,"\n");
            offset += nx * sizeof(float);

            fprintf(fp,"object \"data_mu%d\" class array type float items %d %s ieee data %d\n",
                    number,nx,endian,offset);
            fprintf(fp,"\n");
            offset += nx * sizeof(float);

            number++;
        }
    }

    write_field_dx_fieldlist(fp,potential->blend);

    number = 0;
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            fprintf(fp,"object \"phi%d\" class field\n",number);
            fprintf(fp,"component \"positions\" value \"positions\"\n");
            fprintf(fp,"component \"connections\" value \"connections\"\n");
            fprintf(fp,"component \"data\" value \"data_phi%d\"\n",number);
            fprintf(fp,"\n");

            fprintf(fp,"object \"mu%d\" class field\n",number);
            fprintf(fp,"component \"positions\" value \"positions\"\n");
            fprintf(fp,"component \"connections\" value \"connections\"\n");
            fprintf(fp,"component \"data\" value \"data_mu%d\"\n",number);
            fprintf(fp,"\n");

            number++;
        }
    }

    fprintf(fp,"end\n");

    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < nx;r++)
            {
                work[r] = psi[p][i][r] * psi[p][i][r];
            }
            fwrite(work,sizeof(float),nx,fp);

            for(r = 0;r < nx;r++)
            {
                work[r] = mu[p][i][r];
            }
            fwrite(work,sizeof(float),nx,fp);
        }
    }

    fclose(fp);

    free(work);
}

/*---------------------------------------------------------------------------*/
void write_field_dx_binary_2d(char *filename,density_field *density,
                              chemical_potential *potential)
{
    FILE *fp;
    int p,i,r;
    int number,offset;
    int np,*ns;
    int nx,ny;
    double lx,ly;
    double ***psi,***mu;
    float *work;
    char *endian;

#ifdef WORDS_BIGENDIAN
    endian = "msb";
#else
    endian = "lsb";
#endif

    np = density->blend->np;
    ns = density->blend->ns;
    nx = density->geometry->nx;
    ny = density->geometry->ny;
    lx = density->geometry->lx;
    ly = density->geometry->ly;
    psi = density->psi;
    mu = potential->mu;
    work = xmalloc(sizeof(float) * nx * ny);

    fp = fopen(filename,"w");

    fprintf(fp,"object \"positions\" class gridpositions counts %d %d\n",nx,ny);
    fprintf(fp,"origin\t0\t0\n");
    fprintf(fp,"delta\t%g\t0\n",lx / nx);
    fprintf(fp,"delta\t0\t%g\n",ly / ny);
    fprintf(fp,"\n");

    fprintf(fp,"object \"connections\" class gridconnections counts %d %d\n",nx,ny);
    fprintf(fp,"\n");

    number = 0;
    offset = 0;
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            fprintf(fp,"object \"data_phi%d\" class array type float items %d %s ieee data %d\n",
                    number,nx * ny,endian,offset);
            fprintf(fp,"\n");
            offset += nx * ny * sizeof(float);

            fprintf(fp,"object \"data_mu%d\" class array type float items %d %s ieee data %d\n",
                    number,nx * ny,endian,offset);
            fprintf(fp,"\n");
            offset += nx * ny * sizeof(float);

            number++;
        }
    }

    write_field_dx_fieldlist(fp,potential->blend);

    number = 0;
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            fprintf(fp,"object \"phi%d\" class field\n",number);
            fprintf(fp,"component \"positions\" value \"positions\"\n");
            fprintf(fp,"component \"connections\" value \"connections\"\n");
            fprintf(fp,"component \"data\" value \"data_phi%d\"\n",number);
            fprintf(fp,"\n");

            fprintf(fp,"object \"mu%d\" class field\n",number);
            fprintf(fp,"component \"positions\" value \"positions\"\n");
            fprintf(fp,"component \"connections\" value \"connections\"\n");
            fprintf(fp,"component \"data\" value \"data_mu%d\"\n",number);
            fprintf(fp,"\n");

            number++;
        }
    }

    fprintf(fp,"end\n");

    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < nx * ny;r++)
            {
                work[r] = psi[p][i][r] * psi[p][i][r];
            }
            fwrite(work,sizeof(float),nx * ny,fp);

            for(r = 0;r < nx * ny;r++)
            {
                work[r] = mu[p][i][r];
            }
            fwrite(work,sizeof(float),nx * ny,fp);
        }
    }

    fclose(fp);

    free(work);
}

/*---------------------------------------------------------------------------*/
void write_field_dx_binary_3d(char *filename,density_field *density,
                              chemical_potential *potential)
{
    FILE *fp;
    int p,i,r;
    int number,offset;
    int np,*ns;
    int nx,ny,nz;
    double lx,ly,lz;
    double ***psi,***mu;
    float *work;
    char *endian;

#ifdef WORDS_BIGENDIAN
    endian = "msb";
#else
    endian = "lsb";
#endif

    np = density->blend->np;
    ns = density->blend->ns;
    nx = density->geometry->nx;
    ny = density->geometry->ny;
    nz = density->geometry->nz;
    lx = density->geometry->lx;
    ly = density->geometry->ly;
    lz = density->geometry->lz;
    psi = density->psi;
    mu = potential->mu;
    work = xmalloc(sizeof(float) * nx * ny * nz);

    fp = fopen(filename,"w");

    fprintf(fp,"object \"positions\" class gridpositions counts %d %d %d\n",nx,ny,nz);
    fprintf(fp,"origin\t0\t0\t0\n");
    fprintf(fp,"delta\t%g\t0\t0\n",lx / nx);
    fprintf(fp,"delta\t0\t%g\t0\n",ly / ny);
    fprintf(fp,"delta\t0\t0\t%g\n",lz / nz);
    fprintf(fp,"\n");

    fprintf(fp,"object \"connections\" class gridconnections counts %d %d %d\n",nx,ny,nz);
    fprintf(fp,"\n");

    number = 0;
    offset = 0;
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            fprintf(fp,"object \"data_phi%d\" class array type float items %d %s ieee data %d\n",
                    number,nx * ny * nz,endian,offset);
            fprintf(fp,"\n");
            offset += nx * ny * nz * sizeof(float);

            fprintf(fp,"object \"data_mu%d\" class array type float items %d %s ieee data %d\n",
                    number,nx * ny * nz,endian,offset);
            fprintf(fp,"\n");
            offset += nx * ny * nz * sizeof(float);

            number++;
        }
    }

    write_field_dx_fieldlist(fp,potential->blend);

    number = 0;
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            fprintf(fp,"object \"phi%d\" class field\n",number);
            fprintf(fp,"component \"positions\" value \"positions\"\n");
            fprintf(fp,"component \"connections\" value \"connections\"\n");
            fprintf(fp,"component \"data\" value \"data_phi%d\"\n",number);
            fprintf(fp,"\n");

            fprintf(fp,"object \"mu%d\" class field\n",number);
            fprintf(fp,"component \"positions\" value \"positions\"\n");
            fprintf(fp,"component \"connections\" value \"connections\"\n");
            fprintf(fp,"component \"data\" value \"data_mu%d\"\n",number);
            fprintf(fp,"\n");

            number++;
        }
    }

    fprintf(fp,"end\n");

    for(p = 0;p < np;p++)
    {
        for(i = 0;i < ns[p];i++)
        {
            for(r = 0;r < nx * ny * nz;r++)
            {
                work[r] = psi[p][i][r] * psi[p][i][r];
            }
            fwrite(work,sizeof(float),nx * ny * nz,fp);

            for(r = 0;r < nx * ny * nz;r++)
            {
                work[r] = mu[p][i][r];
            }
            fwrite(work,sizeof(float),nx * ny * nz,fp);
        }
    }

    fclose(fp);

    free(work);
}

/*---------------------------------------------------------------------------*/
void write_field_dx_fieldlist(FILE *fp,polymer_blend *blend)
{
    int i,nc,shape;

    nc = block_number(blend);

        /*  maximum length of "phi%d" */
    shape = 3 + digits_in_decimal(nc) + 1;

    fprintf(fp,"object \"fieldlist\" class array type string rank 1 shape %d items %d data follows\n",
            shape,nc * 2);

    for(i = 0;i < nc;i++)
    {
        fprintf(fp,"\"phi%d\" ",i);
    }
    for(i = 0;i < nc;i++)
    {
        fprintf(fp,"\"mu%d\" ",i);
    }
    fprintf(fp,"\n");
    fprintf(fp,"\n");

}

/*---------------------------------------------------------------------------*/
int digits_in_decimal(int n)
{
    int i;

    i = 0;

    if(n < 0)
    {
        n = -n;
        i++;
    }

    do
    {
        n /= 10;
        i++;
    }
    while(n != 0);

    return i;
}

/*---------------------------------------------------------------------------*/
