/*---------------------------------------------------------------------------*/

/*!  \file  gaussseidel.h
     \brief  Gauss-Seidel relaxation library.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#ifndef INCLUDE_GAUSSSEIDEL_H
#define INCLUDE_GAUSSSEIDEL_H

/*---------------------------------------------------------------------------*/
/*!  \brief  Gauss-Seidel relaxation for 1D field.
     \param  nx  number of division in x axis
     \param  lx  size of the system in x direction
     \param  u  solution of the Poisson equation
     \param  f  source for the Poisson equation  */
void gauss_seidel_1d(int nx,double lx,
                     double *u,double *f);
/*!  \brief  Gauss-Seidel relaxation for 2D field.
     \param  nx  number of division in x axis
     \param  ny  number of division in y axis
     \param  lx  size of the system in x direction
     \param  ly  size of the system in y direction
     \param  u  solution of the Poisson equation
     \param  f  source for the Poisson equation  */
void gauss_seidel_2d(int nx,int ny,
                     double lx,double ly,
                     double *u,double *f);
/*!  \brief  Gauss-Seidel relaxation for 3D field.
     \param  nx  number of division in x axis
     \param  ny  number of division in y axis
     \param  nz  number of division in z axis
     \param  lx  size of the system in x direction
     \param  ly  size of the system in y direction
     \param  lz  size of the system in z direction
     \param  u  solution of the Poisson equation
     \param  f  source for the Poisson equation  */
void gauss_seidel_3d(int nx,int ny,int nz,
                     double lx,double ly,double lz,
                     double *u,double *f);

/*!  \brief  Gauss-Seidel iteration for 1D field.
     \param  n_iteration  number of iteration step
     \param  nx  number of division in x axis
     \param  lx  size of the system in x direction
     \param  u  solution of the Poisson equation
     \param  f  source for the Poisson equation  */
void gauss_seidel_iteration_1d(int n_iteration,
                               int nx,double lx,
                               double *u,double *f);
/*!  \brief  Gauss-Seidel iteration for 2D field.
     \param  n_iteration  number of iteration step
     \param  nx  number of division in x axis
     \param  ny  number of division in y axis
     \param  lx  size of the system in x direction
     \param  ly  size of the system in y direction
     \param  u  solution of the Poisson equation
     \param  f  source for the Poisson equation  */
void gauss_seidel_iteration_2d(int n_iteration,
                               int nx,int ny,
                               double lx,double ly,
                               double *u,double *f);
/*!  \brief  Gauss-Seidel iteration for 3D field.
     \param  n_iteration  number of iteration step
     \param  nx  number of division in x axis
     \param  ny  number of division in y axis
     \param  nz  number of division in z axis
     \param  lx  size of the system in x direction
     \param  ly  size of the system in y direction
     \param  lz  size of the system in z direction
     \param  u  solution of the Poisson equation
     \param  f  source for the Poisson equation  */
void gauss_seidel_iteration_3d(int n_iteration,
                               int nx,int ny,int nz,
                               double lx,double ly,double lz,
                               double *u,double *f);

/*!  \brief  calculate the residual for 1D field.
     \param  nx  number of division in x axis
     \param  lx  size of the system in x direction
     \param  u  solution of the Poisson equation
     \param  f  source for the Poisson equation
     \param  r  residual for the Poisson equation  */
double calculate_residual_1d(int nx,double lx,
                             double *u,double *f,double *r);
/*!  \brief  calculate the residual for 2D field.
     \param  nx  number of division in x axis
     \param  ny  number of division in y axis
     \param  lx  size of the system in x direction
     \param  ly  size of the system in y direction
     \param  u  solution of the Poisson equation
     \param  f  source for the Poisson equation
     \param  r  residual for the Poisson equation  */
double calculate_residual_2d(int nx,int ny,
                             double lx,double ly,
                             double *u,double *f,double *r);
/*!  \brief  calculate the residual for 3D field.
     \param  nx  number of division in x axis
     \param  ny  number of division in y axis
     \param  nz  number of division in z axis
     \param  lx  size of the system in x direction
     \param  ly  size of the system in y direction
     \param  lz  size of the system in z direction
     \param  u  solution of the Poisson equation
     \param  f  source for the Poisson equation
     \param  r  residual for the Poisson equation  */
double calculate_residual_3d(int nx,int ny,int nz,
                             double lx,double ly,double lz,
                             double *u,double *f,double *r);

/*---------------------------------------------------------------------------*/
#endif

/*---------------------------------------------------------------------------*/
