/*---------------------------------------------------------------------------*/

/*!  \file  geometry.c
     \brief  the geometry infomation for the field.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005-2007 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#include "xmalloc.h"
#include "geometry.h"


/*---------------------------------------------------------------------------*/
static void calculate_q2_1d(field_geometry *geometry);
static void calculate_q2_2d(field_geometry *geometry);
static void calculate_q2_3d(field_geometry *geometry);
static void calculate_q2i_1d(field_geometry *geometry);
static void calculate_q2i_2d(field_geometry *geometry);
static void calculate_q2i_3d(field_geometry *geometry);

/*---------------------------------------------------------------------------*/
field_geometry *allocate_field_geometry(void)
{
    field_geometry *geometry;

    geometry = xmalloc(sizeof(field_geometry));

    return geometry;
}

/*---------------------------------------------------------------------------*/
void free_field_geometry(field_geometry *geometry)
{
    if(geometry == NULL)
        return;

    free(geometry);
}

/*---------------------------------------------------------------------------*/
void allocate_q2_q2i(field_geometry *geometry)
{
    int n2;

    n2 = lattice_point_number(geometry);

    geometry->q2 = xmalloc(sizeof(double) * n2);
    geometry->q2i = xmalloc(sizeof(double) * n2);
}

/*---------------------------------------------------------------------------*/
void free_q2_q2i(field_geometry *geometry)
{
    free(geometry->q2);
    free(geometry->q2i);
}

/*---------------------------------------------------------------------------*/
int lattice_point_number(field_geometry *geometry)
{
    int n;

    switch(geometry->dimension)
    {
        case 1:
            n = geometry->nx;
            break;

        case 2:
            n = geometry->nx * geometry->ny;
            break;

        case 3:
            n = geometry->nx * geometry->ny * geometry->nz;
            break;

        default:
            n = 0;
            break;
    }

    return n;
}

/*---------------------------------------------------------------------------*/
int lattice_point_number_complex(field_geometry *geometry)
{
    int n;

    switch(geometry->dimension)
    {
        case 1:
            n = geometry->nx / 2 + 1;
            break;

        case 2:
            n = geometry->nx * (geometry->ny / 2 + 1);
            break;

        case 3:
            n = geometry->nx * geometry->ny * (geometry->nz / 2 + 1);
            break;

        default:
            n = 0;
            break;
    }

    return n;
}

/*---------------------------------------------------------------------------*/
void calculate_q2i(field_geometry *geometry)
{
    switch(geometry->dimension)
    {
        case 1:
            calculate_q2i_1d(geometry);
            break;
        case 2:
            calculate_q2i_2d(geometry);
            break;
        case 3:
            calculate_q2i_3d(geometry);
            break;
        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void calculate_q2(field_geometry *geometry)
{
    switch(geometry->dimension)
    {
        case 1:
            calculate_q2_1d(geometry);
            break;
        case 2:
            calculate_q2_2d(geometry);
            break;
        case 3:
            calculate_q2_3d(geometry);
            break;
        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void calculate_q2i_1d(field_geometry *geometry)
{
    int i, nx, nx2;
    double lx;
    double qx;
    double *q2i;

    lx = geometry->lx;
    q2i = geometry->q2i;
    nx = geometry->nx;
    nx2 = nx / 2 + 1;

    for(i = 0;i < nx2;i++)
    {
        qx = 2 * M_PI * i / lx;

        if(i != 0)
        {
            q2i[i] = 1.0 / (qx * qx);
        }
    }
    q2i[0] = 0.0;
}

/*---------------------------------------------------------------------------*/
void calculate_q2i_2d(field_geometry *geometry)
{
    int i, j, nx, ny, ny2;
    double lx, ly;
    double qx, qy;
    double *q2i;

    lx = geometry->lx;
    ly = geometry->ly;
    q2i = geometry->q2i;
    nx = geometry->nx;
    ny = geometry->ny;
    ny2 = ny / 2 + 1;

    for(i = 0;i < nx;i++)
    {
        if(i < (nx / 2))
            qx = 2 * M_PI * i / lx;
        else
            qx = 2 * M_PI * (i - nx) / lx;

        for(j = 0;j < ny2;j++)
        {
            qy = 2 * M_PI * j / ly;

            if((i != 0) || (j != 0))
            {
                q2i[i * ny2 + j] = 1.0 / (qx * qx + qy * qy);
            }
        }
    }
    q2i[0] = 0.0;
}

/*---------------------------------------------------------------------------*/
void calculate_q2i_3d(field_geometry *geometry)
{
    int i, j, k, nx, ny, nz, nz2;
    double lx, ly, lz;
    double qx, qy, qz;
    double *q2i;

    lx = geometry->lx;
    ly = geometry->ly;
    lz = geometry->lz;
    q2i = geometry->q2i;
    nx = geometry->nx;
    ny = geometry->ny;
    nz = geometry->nz;
    nz2 = nz / 2 + 1;

    for(i = 0;i < nx;i++)
    {
        if(i < (nx / 2))
            qx = 2 * M_PI * i / lx;
        else
            qx = 2 * M_PI * (i - nx) / lx;

        for(j = 0;j < ny;j++)
        {
            if(j < (ny / 2))
                qy = 2 * M_PI * j / ly;
            else
                qy = 2 * M_PI * (j - ny) / ly;

            for(k = 0;k < nz2;k++)
            {
                qz = 2 * M_PI * k / lz;

                if((i != 0) || (j != 0) || (k != 0))
                {
                    q2i[(i * ny + j) * nz2 + k] =
                        1.0 / (qx * qx + qy * qy + qz * qz);
                }
            }
        }
    }
    q2i[0] = 0.0;
}

/*---------------------------------------------------------------------------*/
void calculate_q2_1d(field_geometry *geometry)
{
    int i, nx, nx2;
    double lx;
    double qx;
    double *q2;

    lx = geometry->lx;
    q2 = geometry->q2;
    nx = geometry->nx;
    nx2 = nx / 2 + 1;

    for(i = 0;i < nx2;i++)
    {
        qx = 2 * M_PI * i / lx;

        q2[i] = (qx * qx);
    }
}

/*---------------------------------------------------------------------------*/
void calculate_q2_2d(field_geometry *geometry)
{
    int i, j, nx, ny, ny2;
    double lx, ly;
    double qx, qy;
    double *q2;

    lx = geometry->lx;
    ly = geometry->ly;
    q2 = geometry->q2;
    nx = geometry->nx;
    ny = geometry->ny;
    ny2 = ny / 2 + 1;

    for(i = 0;i < nx;i++)
    {
        if(i < (nx / 2))
            qx = 2 * M_PI * i / lx;
        else
            qx = 2 * M_PI * (i - nx) / lx;

        for(j = 0;j < ny2;j++)
        {
            qy = 2 * M_PI * j / ly;

            q2[i * ny2 + j] = (qx * qx + qy * qy);
        }
    }
}

/*---------------------------------------------------------------------------*/
void calculate_q2_3d(field_geometry *geometry)
{
    int i, j, k, nx, ny, nz, nz2;
    double lx, ly, lz;
    double qx, qy, qz;
    double *q2;

    lx = geometry->lx;
    ly = geometry->ly;
    lz = geometry->lz;
    q2 = geometry->q2;
    nx = geometry->nx;
    ny = geometry->ny;
    nz = geometry->nz;
    nz2 = nz / 2 + 1;

    for(i = 0;i < nx;i++)
    {
        if(i < (nx / 2))
            qx = 2 * M_PI * i / lx;
        else
            qx = 2 * M_PI * (i - nx) / lx;

        for(j = 0;j < ny;j++)
        {
            if(j < (ny / 2))
                qy = 2 * M_PI * j / ly;
            else
                qy = 2 * M_PI * (j - ny) / ly;

            for(k = 0;k < nz2;k++)
            {
                qz = 2 * M_PI * k / lz;

                q2[(i * ny + j) * nz2 + k] =
                    (qx * qx + qy * qy + qz * qz);
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void print_lattice(FILE *fp,field_geometry *geometry)
{
    switch(geometry->dimension)
    {
        case 1:
            fprintf(fp,"lattice: %g\n",geometry->lx);
            break;

        case 2:
            fprintf(fp,"lattice: %g %g\n",geometry->lx,geometry->ly);
            break;

        case 3:
            fprintf(fp,"lattice: %g %g %g\n",geometry->lx,geometry->ly,
                    geometry->lz);
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
int maximum_number_of_division_in_one_axis(field_geometry *geometry)
{
    switch(geometry->dimension)
    {
        case 1:
            return geometry->nx;

        case 2:
            if(geometry->nx > geometry->ny)
            {
                return geometry->nx;
            }
            else
            {
                return geometry->ny;
            }

        case 3:
            if(geometry->nx > geometry->ny)
            {
                if(geometry->nx > geometry->nz)
                {
                    return geometry->nx;
                }
            }
            else
            {
                if(geometry->ny > geometry->nz)
                {
                    return geometry->ny;
                }
            }
            return geometry->nz;

        default:
            break;
    }

    return 0;
}

/*---------------------------------------------------------------------------*/
