/*---------------------------------------------------------------------------*/

/*!  \file  lu.c
     \brief  LU decomposition method library.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005,2006 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include "xmalloc.h"
#include "lu.h"
#include "matrix.h"

static void lu_inverse(double **a,double **b,int n,int *index);

/*---------------------------------------------------------------------------*/
double lu(double **a,int n,int *index)
{
    int i,j,k;
    double u,t;
    double det;
    double *w;

        /*  calculate weighting factor  */
    w = xmalloc(sizeof(double) * n);
    for(i = 0;i < n;i++)
    {
        index[i] = i;
        u = 0.0;

        for(j = 0;j < n;j++)
        {
            t = fabs(a[i][j]);
            if(t > u)
                u = t;
        }

        if(u == 0.0)
            return 0.0;

        w[i] = 1.0 / u;
    }

        /*  LU decomposition method with poviting  */
    det = 1;
    for(k = 0;k < n;k++)
    {
        u = -1.0;
        j = 0.0;
        for(i = k;i < n;i++)
        {
            t = fabs(a[index[i]][k]) * w[index[i]];
            if(t > u)
            {
                u = t;
                j = i;
            }
        }

        if(j != k)
        {
            i = index[j];
            index[j] = index[k];
            index[k] = i;
            det *= -1;
        }

        u = a[index[k]][k];
        det *= u;
        if(u == 0.0)
            return 0.0;
        for(i = (k + 1);i < n;i++)
        {
            a[index[i]][k] /= u;
            t = a[index[i]][k];
            for(j = (k + 1);j < n;j++)
                a[index[i]][j] -= t * a[index[k]][j];
        }
    }

    free(w);

    return det;
}

/*---------------------------------------------------------------------------*/
double inverse(double **a,double **b,int n)
{
    int i,j;
    int *index;
    double det;
    double **c;

    index = xmalloc(sizeof(int *) * n);
    c = allocate_matrix(n);

        /*  since LU decomposition destroy 'a', we use another matrix.  */
    for(i = 0;i < n;i++)
    {
        for(j = 0;j < n;j++)
        {
            c[i][j] = a[i][j];
        }
    }

    det = lu(c,n,index);
    if(det != 0.0)
        lu_inverse(c,b,n,index);

    free(index);
    free_matrix(c);

    return det;
}

/*---------------------------------------------------------------------------*/
void lu_inverse(double **a,double **b,int n,int *index)
{
    int i,j,k;
    double t;
    double *x;

    x = xmalloc(sizeof(double) * n);

        /*  set b to the unit matrix.  */
    for(i = 0;i < n;i++)
    {
        for(j = i;j < n;j++)
        {
            b[i][j] = 0.0;
            b[j][i] = 0.0;
        }
        b[i][i] = 1.0;
    }

        /*  calculate A inverse  */
    for(k = 0;k < n;k++)
    {
        for(i = 0;i < n;i++)
        {
            t = b[index[i]][k];
            for(j = 0;j < i;j++)
                t -= a[index[i]][j] * x[j];
            x[i] = t;
        }
        for(i = (n - 1);i > (-1);i--)
        {
            t = x[i];
            for(j = (i + 1);j < n;j++)
                t -= a[index[i]][j] * x[j];
            x[i] = t / a[index[i]][i];
        }

        for(i = 0;i < n;i++)
            b[i][k] = x[i];
    }

    free(x);
}

/*---------------------------------------------------------------------------*/
