/*---------------------------------------------------------------------------*/

/*!  \file  luaread.c
     \brief  read input Lua file for the simulator  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005-2007 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_LUA_H
#include <lua.h>
#endif
#ifdef HAVE_LUALIB_H
#include <lualib.h>
#endif
#ifdef HAVE_LAUXLIB_H
#include <lauxlib.h>
#endif

#ifndef EXIT_FAILURE
#define EXIT_FAILURE 1
#endif

#include "xmalloc.h"
#include "luautil.h"
#include "luaread.h"
#include "bfs.h"

/*---------------------------------------------------------------------------*/
typedef struct original_polymer_data
{
    int ns;
    double N;
    double *f;
    int **a;
    double lambda;
    char **monomer;
} original_polymer_data;

typedef struct original_monomer_data
{
    int nm;
    char **name;
    double *b;
    double **chi;
} original_monomer_data;

/*---------------------------------------------------------------------------*/
static simulation_condition *read_simulation_condition_lua(lua_State *L);
static simulation_iofiles *read_simulation_iofiles_lua(lua_State *L,
                                                       simulation_condition *condition);
static field_geometry *read_field_geometry_lua(lua_State *L);
static polymer_blend *read_polymer_blend_lua(lua_State *L);
static multigrid_condition *read_multigrid_condition_lua(lua_State *L);

static original_monomer_data *read_original_monomer_data(lua_State *L);
static original_polymer_data *read_original_polymer_data(lua_State *L,
                                                         char *polymer_name);

static original_monomer_data *allocate_original_monomer_data(void);
static void free_original_monomer_data(original_monomer_data *monomer);
static original_polymer_data *allocate_original_polymer_data(void);
static void free_original_polymer_data(original_polymer_data *polymer);

static void correct_matrices(int np,double *volume_fraction,
                             original_monomer_data *monomer,
                             original_polymer_data **polymer);
static polymer_blend *allocate_polymer_blend_using_data(int np,int *ns,double *vf,
                                                        original_monomer_data *monomer,
                                                        original_polymer_data **polymer);
static double get_b(char *name,original_monomer_data *monomer);
static double get_chi(char *name1,char *name2,
                      original_monomer_data *monomer);
static void set_l2(int ns,original_monomer_data *monomer,
                   original_polymer_data *polymer,
                   double **l2);

/*---------------------------------------------------------------------------*/
void read_input_lua(char *input_lua,
                    simulation_condition **condition,
                    simulation_iofiles **iofiles,
                    field_geometry **geometry,
                    polymer_blend **blend,
                    multigrid_condition **mgcondition)
{
    lua_State *L;

    L = luau_open_file(input_lua);
    if(L == NULL)
    {
        exit(EXIT_FAILURE);
    }

    *condition = read_simulation_condition_lua(L);
    *iofiles = read_simulation_iofiles_lua(L,*condition);
    *geometry = read_field_geometry_lua(L);
    *blend = read_polymer_blend_lua(L);
    if((*condition)->use_multigrid_solver)
        *mgcondition = read_multigrid_condition_lua(L);
    else
        *mgcondition = NULL;

    lua_close(L);
}

/*---------------------------------------------------------------------------*/
simulation_condition *read_simulation_condition_lua(lua_State *L)
{
    simulation_condition *condition;

    condition = allocate_simulation_condition();


    luau_push_table_or_exit(L,"condition");


    luau_get_bool_or_exit(L,"optimize_lattice",&condition->optimize_lattice);
    luau_get_bool_or_exit(L,"use_multigrid_solver",&condition->use_multigrid_solver);
    luau_get_bool_or_exit(L,"dynamics_simulation",&condition->dynamics_simulation);
    luau_get_bool_or_exit(L,"save_psi_sequential",&condition->save_psi_sequential);
    luau_get_bool_or_exit(L,"save_density_sequential",&condition->save_density_sequential);
    luau_get_bool_or_exit(L,"save_chemical_potential_sequential",&condition->save_chemical_potential_sequential);
    luau_get_bool_or_exit(L,"save_free_energy_sequential",&condition->save_free_energy_sequential);
    luau_get_bool_or_exit(L,"save_geometry_sequential",&condition->save_geometry_sequential);
    luau_get_bool_or_exit(L,"save_dx_sequential",&condition->save_dx_sequential);

    luau_get_double_or_exit(L,"error_tolerance",&condition->error_tolerance);
    luau_get_unsigned_long_or_exit(L,"seed",&condition->seed);
    luau_get_double_or_exit(L,"noise",&condition->noise);
    luau_get_double_or_exit(L,"initial_noise",&condition->initial_noise);
    luau_get_double_or_exit(L,"phi_min",&condition->phi_min);

    luau_get_int_or_exit(L,"iteration_max",&condition->iteration_max);
    luau_get_int_or_exit(L,"interval",&condition->interval);
    luau_get_double_or_exit(L,"omega",&condition->omega);


    luau_pop_table_or_exit(L);


    return condition;
}

/*---------------------------------------------------------------------------*/
simulation_iofiles *read_simulation_iofiles_lua(lua_State *L,
                                                simulation_condition *condition)
{
    simulation_iofiles *iofiles;

    iofiles = allocate_simulation_iofiles();


    luau_push_table_or_exit(L,"file");


    luau_get_string_or_exit(L,"wisdom",&iofiles->wisdom);

    if(condition->save_psi_sequential)
        luau_get_string_or_exit(L,"psi_template",&iofiles->psi_template);
    else
        luau_get_string_or_exit(L,"psi_output",&iofiles->psi_output);

    if(condition->save_density_sequential)
        luau_get_string_or_exit(L,"density_template",&iofiles->density_template);
    else
        luau_get_string_or_exit(L,"density_output",&iofiles->density_output);

    if(condition->save_chemical_potential_sequential)
        luau_get_string_or_exit(L,"chemical_potential_template",&iofiles->chemical_potential_template);
    else
        luau_get_string_or_exit(L,"chemical_potential_output",&iofiles->chemical_potential_output);

    if(condition->save_free_energy_sequential)
        luau_get_string_or_exit(L,"free_energy_template",&iofiles->free_energy_template);
    else
        luau_get_string_or_exit(L,"free_energy_output",&iofiles->free_energy_output);

    if(condition->save_geometry_sequential)
        luau_get_string_or_exit(L,"geometry_template",&iofiles->geometry_template);
    else
        luau_get_string_or_exit(L,"geometry_output",&iofiles->geometry_output);

    if(condition->save_dx_sequential)
        luau_get_string_or_exit(L,"dx_template",&iofiles->dx_template);
    else
        luau_get_string_or_exit(L,"dx_output",&iofiles->dx_output);


    luau_pop_table_or_exit(L);


    return iofiles;
}

/*---------------------------------------------------------------------------*/
field_geometry *read_field_geometry_lua(lua_State *L)
{
    field_geometry *geometry;

    geometry = allocate_field_geometry();


    luau_push_table_or_exit(L,"geometry");


    luau_get_int_or_exit(L,"dimension",&geometry->dimension);

    if(geometry->dimension >= 1)
    {
        luau_get_int_or_exit(L,"nx",&geometry->nx);
        luau_get_double_or_exit(L,"lx",&geometry->lx);
    }
    if(geometry->dimension >= 2)
    {
        luau_get_int_or_exit(L,"ny",&geometry->ny);
        luau_get_double_or_exit(L,"ly",&geometry->ly);
    }
    if(geometry->dimension >= 3)
    {
        luau_get_int_or_exit(L,"nz",&geometry->nz);
        luau_get_double_or_exit(L,"lz",&geometry->lz);
    }


    luau_pop_table_or_exit(L);


    return geometry;
}

/*---------------------------------------------------------------------------*/
polymer_blend *read_polymer_blend_lua(lua_State *L)
{
    polymer_blend *blend;
    char **polymer_name;
    double *volume_fraction;
    int np,*ns;
    int i;
    original_monomer_data *monomer;
    original_polymer_data **polymer;

    luau_push_table_or_exit(L,"blend");


    np = luau_get_string_array_or_exit(L,"polymer",&polymer_name);
    luau_get_double_array_or_exit(L,"volume_fraction",&volume_fraction);


    luau_pop_table_or_exit(L);



    monomer = read_original_monomer_data(L);

    polymer = xmalloc(sizeof(original_polymer_data *) * np);
    ns = xmalloc(sizeof(int) * np);
    for(i = 0;i < np;i++)
    {
        polymer[i] = read_original_polymer_data(L,polymer_name[i]);
        ns[i] = polymer[i]->ns;
    }


    correct_matrices(np,volume_fraction,monomer,polymer);


        /*  set the blend data from the monomer data and the polymer data.  */
    blend = allocate_polymer_blend_using_data(np,ns,volume_fraction,monomer,polymer);


    free_original_monomer_data(monomer);
    for(i = 0;i < np;i++)
    {
        free_original_polymer_data(polymer[i]);
    }
    free(polymer);
    free(ns);


    return blend;
}

/*---------------------------------------------------------------------------*/
static multigrid_condition *read_multigrid_condition_lua(lua_State *L)
{
    multigrid_condition *mgcondition;

    mgcondition = allocate_multigrid_condition();


    luau_push_table_or_exit(L,"multigrid");


    luau_get_int_or_exit(L,"n_cycle",&mgcondition->n_cycle);
    luau_get_int_or_exit(L,"n_pre",&mgcondition->n_pre);
    luau_get_int_or_exit(L,"n_post",&mgcondition->n_post);
    luau_get_double_or_exit(L,"error_tolerance",&mgcondition->error_tolerance);


    luau_pop_table_or_exit(L);

    return mgcondition;
}

/*---------------------------------------------------------------------------*/
original_monomer_data *read_original_monomer_data(lua_State *L)
{
    original_monomer_data *monomer;

    monomer = allocate_original_monomer_data();


    luau_push_table_or_exit(L,"monomer");


    monomer->nm = luau_get_string_array_or_exit(L,"name",&monomer->name);
    luau_get_double_array_or_exit(L,"b",&monomer->b);
    luau_get_double_array2d_or_exit(L,"chi",&monomer->chi);


    luau_pop_table_or_exit(L);


    return monomer;
}

/*---------------------------------------------------------------------------*/
static original_polymer_data *read_original_polymer_data(lua_State *L,
                                                         char *polymer_name)
{
    original_polymer_data *polymer;

    polymer = allocate_original_polymer_data();


    luau_push_table_or_exit(L,polymer_name);


    luau_get_double_or_exit(L,"N",&polymer->N);
    polymer->ns = luau_get_string_array_or_exit(L,"monomer",&polymer->monomer);
    luau_get_double_array_or_exit(L,"f",&polymer->f);
    luau_get_bool_array2d_or_exit(L,"a",&polymer->a);
    luau_get_double(L,"lambda",&polymer->lambda);

    luau_pop_table_or_exit(L);

    return polymer;
}

/*---------------------------------------------------------------------------*/
original_monomer_data *allocate_original_monomer_data(void)
{
    original_monomer_data *monomer;

    monomer = xmalloc(sizeof(original_monomer_data));
    monomer->name = NULL;
    monomer->b = NULL;
    monomer->chi = NULL;

    return monomer;
}

/*---------------------------------------------------------------------------*/
void free_original_monomer_data(original_monomer_data *monomer)
{
    luau_free_string_array(monomer->name);
    luau_free_double_array(monomer->b);
    luau_free_double_array2d(monomer->chi);
    free(monomer);
}

/*---------------------------------------------------------------------------*/
original_polymer_data *allocate_original_polymer_data(void)
{
    original_polymer_data *polymer;

    polymer = xmalloc(sizeof(original_polymer_data));

    polymer->f = NULL;
    polymer->a = NULL;
    polymer->monomer = NULL;

    return polymer;
}

/*---------------------------------------------------------------------------*/
void free_original_polymer_data(original_polymer_data *polymer)
{
    luau_free_double_array(polymer->f);
    luau_free_bool_array2d(polymer->a);
    luau_free_string_array(polymer->monomer);
    free(polymer);
}

/*---------------------------------------------------------------------------*/
void correct_matrices(int np,double *volume_fraction,
                      original_monomer_data *monomer,
                      original_polymer_data **polymer)
{
    int p,i,j;
    double vf_tot,f_tot;

        /*  Correct volume fraction.  */
    vf_tot = 0;
    for(p = 0;p < np;p++)
        vf_tot += volume_fraction[p];
    for(p = 0;p < np;p++)
        volume_fraction[p] /= vf_tot;

        /*  Correct block ratio.  */
    for(p = 0;p < np;p++)
    {
        f_tot = 0;
        for(i = 0;i < polymer[p]->ns;i++)
            f_tot += polymer[p]->f[i];
        for(i = 0;i < polymer[p]->ns;i++)
            polymer[p]->f[i] /= f_tot;
    }

        /*  Correct adjacent matrix.  */
    for(p = 0;p < np;p++)
    {
        for(i = 0;i < polymer[p]->ns;i++)
        {
            polymer[p]->a[i][i] = TRUE;
            for(j = 0;j < i;j++)
                polymer[p]->a[i][j] = polymer[p]->a[j][i];
        }
    }

        /*  Correct chi parameter.  */
    for(i = 0;i < monomer->nm;i++)
    {
        for(j = 0;j < i;j++)
        {
            monomer->chi[i][j] = monomer->chi[j][i];
        }
    }
}

/*---------------------------------------------------------------------------*/
polymer_blend *allocate_polymer_blend_using_data(int np,int *ns,double *volume_fraction,
                                                 original_monomer_data *monomer,
                                                 original_polymer_data **polymer)
{
    polymer_blend *blend;
    polymer_chain *chain;
    int p,i,q,j;

    blend = allocate_polymer_blend(np,ns);

    for(p = 0;p < np;p++)
    {
        blend->volume_fraction[p] = volume_fraction[p];
        chain = blend->chain[p];
        chain->N = polymer[p]->N;
        chain->lambda = polymer[p]->lambda;
        set_l2(ns[p],monomer,polymer[p],chain->l2);

        for(i = 0;i < ns[p];i++)
        {
            chain->f[i] = polymer[p]->f[i];
            chain->b[i] = get_b(polymer[p]->monomer[i],monomer);

            for(q = 0;q < np;q++)
            {
                for(j = 0;j < ns[q];j++)
                {
                    blend->chi[p][i][q][j] = get_chi(polymer[p]->monomer[i],
                                                     polymer[q]->monomer[j],
                                                     monomer);
                }
            }
        }
    }

    return blend;
}

/*---------------------------------------------------------------------------*/
double get_b(char *name,original_monomer_data *monomer)
{
    int i;

    for(i = 0;i < monomer->nm;i++)
    {
        if(strcmp(name,monomer->name[i]) == 0)
            return monomer->b[i];
    }

    return 0.0;
}

/*---------------------------------------------------------------------------*/
double get_chi(char *name1,char *name2,
               original_monomer_data *monomer)
{
    int i,j;

    for(i = 0;i < monomer->nm;i++)
    {
        if(strcmp(name1,monomer->name[i]) == 0)
        {
            for(j = 0;j < monomer->nm;j++)
            {
                if(strcmp(name2,monomer->name[j]) == 0)
                {
                    return monomer->chi[i][j];
                }
            }
        }
    }

    return 0.0;
}

/*---------------------------------------------------------------------------*/
void set_l2(int ns,original_monomer_data *monomer,
            original_polymer_data *polymer,
            double **l2)
{
    int i;
    double b;
    double *w;

    w = xmalloc(sizeof(double) * ns);

    for(i = 0;i < ns;i++)
    {
        b = get_b(polymer->monomer[i],monomer);
        w[i] = polymer->N * polymer->f[i] * b * b;
    }

    breadth_first_search(ns, polymer->a, l2, w);

    free(w);
}

/*---------------------------------------------------------------------------*/
