/*---------------------------------------------------------------------------*/

/*!  \file  mgpoisson.c
     \brief  the multigrid Poisson equation solver.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005,2006 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include "xmalloc.h"
#include "regrid.h"
#include "mgpoisson.h"
#include "gaussseidel.h"

/*---------------------------------------------------------------------------*/
static void restrict_residual(multigrid_field *multigrid,int level);
static void prolongate_solution(multigrid_field *multigrid,int level);
static void neutralize_source(multigrid_field *multigrid,int level);
static void clear_solution(multigrid_field *multigrid,int level);
static void correct_field(multigrid_field *multigrid,int level);
static void gauss_seidel_iteration(int n,multigrid_field *multigrid,int level);
static double calculate_residual(multigrid_field *multigrid,int level);

/*---------------------------------------------------------------------------*/
multigrid_condition *allocate_multigrid_condition(void)
{
    multigrid_condition *mgcondition;

    mgcondition = xmalloc(sizeof(multigrid_condition));

    return mgcondition;
}

/*---------------------------------------------------------------------------*/
void free_multigrid_condition(multigrid_condition *mgcondition)
{
    free(mgcondition);
}

/*---------------------------------------------------------------------------*/
int operate_Li_multigrid(multigrid_field *multigrid,
                         multigrid_condition *mgcondition,
                         double *psi,double *Li_psi)
{
    int i,r,l;
    int n;
    double Li_psi_bar;
    double r2;
    double error_tolerance;
    int n_cycle,n_pre,n_post;

    error_tolerance = mgcondition->error_tolerance;
    n_cycle = mgcondition->n_cycle;
    n_pre = mgcondition->n_pre;
    n_post = mgcondition->n_post;

    n = lattice_point_number_multigrid(multigrid,0);

        /*  set source field and initial value of approximate solution  */
    for(r = 0;r < n;r++)
    {
        multigrid->f[0][r] = -psi[r];
        multigrid->u[0][r] = Li_psi[r];
    }
    neutralize_source(multigrid,0);

    for(i = 0;i < n_cycle;i++)
    {
        r2 = calculate_residual(multigrid,0);

        if(r2 < error_tolerance)
            break;

            /*  multigrid V-cycle.  */

            /*  to coarser grids.  */
        for(l = 0;l < multigrid->level - 1;l++)
        {
            neutralize_source(multigrid,l);

            gauss_seidel_iteration(n_pre,multigrid,l);

            calculate_residual(multigrid,l);
            restrict_residual(multigrid,l);
            clear_solution(multigrid,l + 1);
        }

            /*  the coarsest grid.
                this grid contains only one lattice point.  */
        multigrid->u[multigrid->level - 1][0] = 0;
        multigrid->v[multigrid->level - 1][0] = 0;
        multigrid->f[multigrid->level - 1][0] = 0;
        multigrid->r[multigrid->level - 1][0] = 0;

            /*  to finer grids.  */
        for(l = multigrid->level - 2;l >= 0;l--)
        {
            prolongate_solution(multigrid,l + 1);
            correct_field(multigrid,l);

            gauss_seidel_iteration(n_post,multigrid,l);
        }
    }

    Li_psi_bar = 0;
    for(r = 0;r < n;r++)
    {
        Li_psi_bar += multigrid->u[0][r];
    }
    Li_psi_bar /= n;

        /*  copy the approximate solution  */
    for(r = 0;r < n;r++)
    {
        Li_psi[r] = multigrid->u[0][r] - Li_psi_bar;
    }

    return i;
}

/*---------------------------------------------------------------------------*/
void restrict_residual(multigrid_field *multigrid,int level)
{
    regrid_field(multigrid,level,level + 1,multigrid->r[level],multigrid->f[level + 1]);
}

/*---------------------------------------------------------------------------*/
void prolongate_solution(multigrid_field *multigrid,int level)
{
    regrid_field(multigrid,level,level - 1,multigrid->u[level],multigrid->v[level - 1]);
}

/*---------------------------------------------------------------------------*/
void neutralize_source(multigrid_field *multigrid,int level)
{
    int r,n;
    double f_bar;

    n = lattice_point_number_multigrid(multigrid,level);

    f_bar = 0;
    for(r = 0;r < n;r++)
    {
        f_bar += multigrid->f[level][r];
    }
    f_bar /= n;

    for(r = 0;r < n;r++)
    {
        multigrid->f[level][r] -= f_bar;
    }
}

/*---------------------------------------------------------------------------*/
void clear_solution(multigrid_field *multigrid,int level)
{
    int r;

    for(r = 0;r < lattice_point_number_multigrid(multigrid,level);r++)
    {
        multigrid->u[level][r] = 0;
    }
}

/*---------------------------------------------------------------------------*/
void correct_field(multigrid_field *multigrid,int level)
{
    int r;

    for(r = 0;r < lattice_point_number_multigrid(multigrid,level);r++)
    {
        multigrid->u[level][r] += multigrid->v[level][r];
    }
}

/*---------------------------------------------------------------------------*/
void gauss_seidel_iteration(int n,multigrid_field *multigrid,int level)
{
    switch(multigrid->dimension)
    {
        case 1:
            gauss_seidel_iteration_1d(n,multigrid->nx[level],multigrid->lx,
                                      multigrid->u[level],multigrid->f[level]);
            break;

        case 2:
            gauss_seidel_iteration_2d(n,multigrid->nx[level],multigrid->ny[level],
                                      multigrid->lx,multigrid->ly,
                                      multigrid->u[level],multigrid->f[level]);
            break;

        case 3:
            gauss_seidel_iteration_3d(n,multigrid->nx[level],multigrid->ny[level],multigrid->nz[level],
                                      multigrid->lx,multigrid->ly,multigrid->lz,
                                      multigrid->u[level],multigrid->f[level]);
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
double calculate_residual(multigrid_field *multigrid,int level)
{
    double r2;

    switch(multigrid->dimension)
    {
        case 1:
            r2 = calculate_residual_1d(multigrid->nx[level],multigrid->lx,
                                       multigrid->u[level],multigrid->f[level],
                                       multigrid->r[level]);
            break;

        case 2:
            r2 = calculate_residual_2d(multigrid->nx[level],multigrid->ny[level],
                                       multigrid->lx,multigrid->ly,
                                       multigrid->u[level],multigrid->f[level],
                                       multigrid->r[level]);
            break;

        case 3:
            r2 = calculate_residual_3d(multigrid->nx[level],multigrid->ny[level],multigrid->nz[level],
                                       multigrid->lx,multigrid->ly,multigrid->lz,
                                       multigrid->u[level],multigrid->f[level],
                                       multigrid->r[level]);
            break;

        default:
            r2 = 0;
            break;
    }

    return r2;
}

/*---------------------------------------------------------------------------*/
