/*---------------------------------------------------------------------------*/

/*!  \file  operator.c
     \brief  perform the field operator to the field  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005-2007 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#ifdef HAVE_FFTW3_H
#ifdef malloc
#undef malloc
#include <fftw3.h>
#define malloc rpl_malloc
#else
#include <fftw3.h>
#endif
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include "xmalloc.h"
#include "operator.h"
#include "ln.h"
#include "laplacian.h"

#define OPERATOR_FFTW_FLAGS (FFTW_EXHAUSTIVE)

/*---------------------------------------------------------------------------*/
static void create_field_fftw_plan_1d(field_fftw_plan *plan);
static void create_field_fftw_plan_2d(field_fftw_plan *plan);
static void create_field_fftw_plan_3d(field_fftw_plan *plan);

/*---------------------------------------------------------------------------*/
field_fftw_plan *allocate_field_fftw_plan(field_geometry *geometry,
                                          char *wisdom_file)
{
    int n;
    int n2;
    field_fftw_plan *plan;
    FILE *fp;


    n = lattice_point_number(geometry);
    n2 = lattice_point_number_complex(geometry);

    plan = xmalloc(sizeof(field_fftw_plan));

    plan->geometry = geometry;
    plan->work = fftw_malloc(sizeof(double) * n);
    plan->F_work = fftw_malloc(sizeof(fftw_complex) * n2);


        /*  import the words of wisdom.  */
    if(strlen(wisdom_file) > 0)
    {
        if((fp = fopen(wisdom_file,"r")) != NULL)
        {
            fftw_import_wisdom_from_file(fp);
            fclose(fp);
        }
    }

    switch(geometry->dimension)
    {
        case 1:
            create_field_fftw_plan_1d(plan);
            break;
        case 2:
            create_field_fftw_plan_2d(plan);
            break;
        case 3:
            create_field_fftw_plan_3d(plan);
            break;
        default:
            break;
    }

        /*  export the words of wisdom.  */
    if(strlen(wisdom_file) > 0)
    {
        if((fp = fopen(wisdom_file,"w")) != NULL)
        {
            fftw_export_wisdom_to_file(fp);
            fclose(fp);
        }
    }


    calculate_q2(geometry);
    calculate_q2i(geometry);

    return plan;
}

/*---------------------------------------------------------------------------*/
static void create_field_fftw_plan_1d(field_fftw_plan *plan)
{
    int nx;

    nx = plan->geometry->nx;

    plan->plan_forward =
        fftw_plan_dft_r2c_1d(nx,
                             plan->work,
                             plan->F_work,
                             OPERATOR_FFTW_FLAGS);
    plan->plan_backward =
        fftw_plan_dft_c2r_1d(nx,
                             plan->F_work,
                             plan->work,
                             OPERATOR_FFTW_FLAGS);
}

/*---------------------------------------------------------------------------*/
static void create_field_fftw_plan_2d(field_fftw_plan *plan)
{
    int nx,ny;

    nx = plan->geometry->nx;
    ny = plan->geometry->ny;

    plan->plan_forward =
        fftw_plan_dft_r2c_2d(nx,ny,
                             plan->work,
                             plan->F_work,
                             OPERATOR_FFTW_FLAGS);
    plan->plan_backward =
        fftw_plan_dft_c2r_2d(nx,ny,
                             plan->F_work,
                             plan->work,
                             OPERATOR_FFTW_FLAGS);
}

/*---------------------------------------------------------------------------*/
static void create_field_fftw_plan_3d(field_fftw_plan *plan)
{
    int nx,ny,nz;

    nx = plan->geometry->nx;
    ny = plan->geometry->ny;
    nz = plan->geometry->nz;

    plan->plan_forward =
        fftw_plan_dft_r2c_3d(nx,ny,nz,
                             plan->work,
                             plan->F_work,
                             OPERATOR_FFTW_FLAGS);
    plan->plan_backward =
        fftw_plan_dft_c2r_3d(nx,ny,nz,
                             plan->F_work,
                             plan->work,
                             OPERATOR_FFTW_FLAGS);
}

/*---------------------------------------------------------------------------*/
void free_field_fftw_plan(field_fftw_plan *plan)
{
    fftw_free(plan->work);
    fftw_free(plan->F_work);

    free(plan);
}

/*---------------------------------------------------------------------------*/
void operate_L(field_geometry *geometry,double *psi,double *L_psi)
{
    switch(geometry->dimension)
    {
        case 1:
            operate_L_1d(geometry,psi,L_psi);
            break;

        case 2:
            operate_L_2d(geometry,psi,L_psi);
            break;

        case 3:
            operate_L_3d(geometry,psi,L_psi);
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void operate_Li(field_geometry *geometry,double *psi,double *Li_psi,
                field_fftw_plan *plan)
{
    int i, n, n2;
    double *work;
    fftw_complex *F_work;
    double *q2i;

    work = plan->work;
    F_work = plan->F_work;
    q2i = geometry->q2i;

    n = lattice_point_number(geometry);
    n2 = lattice_point_number_complex(geometry);

        /*  Copy psi to work.  */
    for(i = 0;i < n;i++)
    {
        work[i] = psi[i];
    }

    fftw_execute(plan->plan_forward);

        /*  Integral operator.  */
    for(i = 0;i < n2;i++)
    {
        c_re(F_work[i]) *= -q2i[i];
        c_im(F_work[i]) *= -q2i[i];
    }

    fftw_execute(plan->plan_backward);

        /*  Copy work to Li_psi and normalize Li_psi.  */
    for(i = 0;i < n;i++)
    {
        Li_psi[i] = work[i] / n;
    }
}

/*---------------------------------------------------------------------------*/
void operate_Li_cut(field_geometry *geometry,double *psi,double *Li_psi,
                    double lambda,field_fftw_plan *plan)
{
    int i, n, n2;
    double *work;
    fftw_complex *F_work;
    double *q2;
    double q2cut;

    work = plan->work;
    F_work = plan->F_work;
    q2 = geometry->q2;
    q2cut = 1.0 / (lambda * lambda);

    n = lattice_point_number(geometry);
    n2 = lattice_point_number_complex(geometry);

        /*  Copy psi to work.  */
    for(i = 0;i < n;i++)
    {
        work[i] = psi[i];
    }

    fftw_execute(plan->plan_forward);

        /*  Integral operator.  */
    for(i = 0;i < n2;i++)
    {
        c_re(F_work[i]) /= -(q2[i] + q2cut);
        c_im(F_work[i]) /= -(q2[i] + q2cut);
    }

    fftw_execute(plan->plan_backward);

        /*  Copy work to Li_psi and normalize Li_psi.  */
    for(i = 0;i < n;i++)
    {
        Li_psi[i] = work[i] / n;
    }
}

/*---------------------------------------------------------------------------*/
void operate_Li_cut_for_statics(field_geometry *geometry,double *psi,double *Li_psi,
                                double lambda,field_fftw_plan *plan)
{
    int i, n, n2;
    double *work;
    fftw_complex *F_work;
    double *q2;
    double q2cut;

    work = plan->work;
    F_work = plan->F_work;
    q2 = geometry->q2;
    q2cut = 1.0 / (lambda * lambda);

    n = lattice_point_number(geometry);
    n2 = lattice_point_number_complex(geometry);

        /*  Copy psi to work.  */
    for(i = 0;i < n;i++)
    {
        work[i] = psi[i];
    }

    fftw_execute(plan->plan_forward);

        /*  Integral operator.  */
    for(i = 1;i < n2;i++)
    {
        c_re(F_work[i]) /= -(q2[i] + q2cut);
        c_im(F_work[i]) /= -(q2[i] + q2cut);
    }
        /*  reset the offset value to 0.  */
    c_re(F_work[0]) = 0;
    c_im(F_work[0]) = 0;

    fftw_execute(plan->plan_backward);

        /*  Copy work to Li_psi and normalize Li_psi.  */
    for(i = 0;i < n;i++)
    {
        Li_psi[i] = work[i] / n;
    }
}

/*---------------------------------------------------------------------------*/
void operate_F(field_geometry *geometry,double *psi,fftw_complex *F_psi,
               field_fftw_plan *plan)
{
    int i, n, n2;
    double *work;
    fftw_complex *F_work;

    work = plan->work;
    F_work = plan->F_work;

    n = lattice_point_number(geometry);
    n2 = lattice_point_number_complex(geometry);

        /*  Copy psi to work.  */
    for(i = 0;i < n;i++)
    {
        work[i] = psi[i];
    }

    fftw_execute(plan->plan_forward);

        /*  Normalize F_work and copy it to F_psi.  */
    for(i = 0;i < n2;i++)
    {
        c_re(F_psi[i]) = c_re(F_work[i]) / n;
        c_im(F_psi[i]) = c_im(F_work[i]) / n;
    }
}

/*---------------------------------------------------------------------------*/
void operate_Fi(field_geometry *geometry,fftw_complex *F_psi,double *psi,
                field_fftw_plan *plan)
{
    int i, n, n2;
    double *work;
    fftw_complex *F_work;

    work = plan->work;
    F_work = plan->F_work;

    n = lattice_point_number(geometry);
    n2 = lattice_point_number_complex(geometry);

        /*  Copy F_psi to F_work.  */
    for(i = 0;i < n2;i++)
    {
        c_re(F_work[i]) = c_re(F_psi[i]);
        c_im(F_work[i]) = c_im(F_psi[i]);
    }

    fftw_execute(plan->plan_backward);

        /*  Copy work to psi.  */
    for(i = 0;i < n;i++)
    {
        psi[i] = work[i];
    }
}

/*---------------------------------------------------------------------------*/
