/*---------------------------------------------------------------------------*/

/*!  \file  regrid.c
     \brief  library for regridding the field.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005,2006 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include "regrid.h"

/*---------------------------------------------------------------------------*/
static double interpolate_1d(double u_l,double u_u,
                             double dx);
static double interpolate_2d(double u_ll,double u_ul,
                             double u_lu,double u_uu,
                             double dx,double dy);
static double interpolate_3d(double u_lll,double u_ull,
                             double u_lul,double u_uul,
                             double u_llu,double u_ulu,
                             double u_luu,double u_uuu,
                             double dx,double dy,double dz);
static void regrid_1d(int nx1,int nx2,
                      double *u1,double *u2);
static void regrid_2d(int nx1,int ny1,
                      int nx2,int ny2,
                      double *u1,double *u2);
static void regrid_3d(int nx1,int ny1,int nz1,
                      int nx2,int ny2,int nz2,
                      double *u1,double *u2);

/*---------------------------------------------------------------------------*/
void regrid_field(multigrid_field *multigrid,int level1,int level2,
                  double *u1,double *u2)
{
    switch(multigrid->dimension)
    {
        case 1:
            regrid_1d(multigrid->nx[level1],multigrid->nx[level2],
                      u1,u2);
            break;

        case 2:
            regrid_2d(multigrid->nx[level1],multigrid->ny[level1],
                      multigrid->nx[level2],multigrid->ny[level2],
                      u1,u2);
            break;

        case 3:
            regrid_3d(multigrid->nx[level1],multigrid->ny[level1],multigrid->nz[level1],
                      multigrid->nx[level2],multigrid->ny[level2],multigrid->nz[level2],
                      u1,u2);
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
double interpolate_1d(double u_l,double u_u,double dx)
{
    double u;

    u = u_l * dx + u_u * (1 - dx);

    return u;
}

/*---------------------------------------------------------------------------*/
double interpolate_2d(double u_ll,double u_ul,double u_lu,double u_uu,
                      double dx,double dy)
{
    double u;

    u = (u_ll * dx + u_ul * (1 - dx)) * dy
        + (u_lu * dx + u_uu * (1 - dx)) * (1 - dy);

    return u;
}

/*---------------------------------------------------------------------------*/
double interpolate_3d(double u_lll,double u_ull,double u_lul,double u_uul,
                      double u_llu,double u_ulu,double u_luu,double u_uuu,
                      double dx,double dy,double dz)
{
    double u;

    u = ((u_lll * dx + u_ull * (1 - dx)) * dy
         + (u_lul * dx + u_uul * (1 - dx)) * (1 - dy)) * dz
        + ((u_llu * dx + u_ulu * (1 - dx)) * dy
           + (u_luu * dx + u_uuu * (1 - dx)) * (1 - dy)) * (1 - dz);

    return u;
}

/*---------------------------------------------------------------------------*/
void regrid_1d(int nx1,int nx2,
               double *u1,double *u2)
{
    int i2;
    int iu,il;
    double x0;
    double dx;

    for(i2 = 0;i2 < nx2;i2++)
    {
        x0 = (double)i2 * nx1 / nx2;
        il = (int)x0;
        dx = x0 - il;
        iu = (il + 1) % nx1;

        u2[i2] = interpolate_1d(u1[il],u1[iu],dx);
    }
}

/*---------------------------------------------------------------------------*/
void regrid_2d(int nx1,int ny1,
               int nx2,int ny2,
               double *u1,double *u2)
{
    int i2,j2;
    int iu,il;
    int ju,jl;
    double x0,y0;
    double dx,dy;

    for(i2 = 0;i2 < nx2;i2++)
    {
        x0 = (double)i2 * nx1 / nx2;
        il = (int)x0;
        dx = x0 - il;
        iu = (il + 1) % nx1;

        for(j2 = 0;j2 < ny2;j2++)
        {
            y0 = (double)j2 * ny1 / ny2;
            jl = (int)y0;
            dy = y0 - jl;
            ju = (jl + 1) % ny1;

            u2[i2 * ny2 + j2] =
                interpolate_2d(u1[il * ny1 + jl],
                               u1[iu * ny1 + jl],
                               u1[il * ny1 + ju],
                               u1[iu * ny1 + ju],
                               dx,dy);
        }
    }
}

/*---------------------------------------------------------------------------*/
void regrid_3d(int nx1,int ny1,int nz1,
               int nx2,int ny2,int nz2,
               double *u1,double *u2)
{
    int i2,j2,k2;
    int iu,il;
    int ju,jl;
    int ku,kl;
    double x0,y0,z0;
    double dx,dy,dz;

    for(i2 = 0;i2 < nx2;i2++)
    {
        x0 = (double)i2 * nx1 / nx2;
        il = (int)x0;
        dx = x0 - il;
        iu = (il + 1) % nx1;

        for(j2 = 0;j2 < ny2;j2++)
        {
            y0 = (double)j2 * ny1 / ny2;
            jl = (int)y0;
            dy = y0 - jl;
            ju = (jl + 1) % ny1;

            for(k2 = 0;k2 < nz2;k2++)
            {
                z0 = (double)k2 * nz1 / nz2;
                kl = (int)z0;
                dz = z0 - kl;
                ku = (kl + 1) % nz1;

                u2[(i2 * ny2 + j2) * nz2 + k2] =
                    interpolate_3d(u1[(il * ny1 + jl) * nz1 + kl],
                                   u1[(iu * ny1 + jl) * nz1 + kl],
                                   u1[(il * ny1 + ju) * nz1 + kl],
                                   u1[(iu * ny1 + ju) * nz1 + kl],
                                   u1[(il * ny1 + jl) * nz1 + ku],
                                   u1[(iu * ny1 + jl) * nz1 + ku],
                                   u1[(il * ny1 + ju) * nz1 + ku],
                                   u1[(iu * ny1 + ju) * nz1 + ku],
                                   dx,dy,dz);
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
