/*---------------------------------------------------------------------------*/

/*!  \file  simulator.c
     \brief  a statics / dynamics simulator.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005-2007 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include "lu.h"
#include "matrix.h"
#include "tridiag.h"
#include "mt.h"
#include "nrand.h"
#include "gpwrite.h"
#include "dxwrite.h"
#include "textwrite.h"
#include "condition.h"
#include "simulator.h"
#include "freeenergy.h"
#include "evolver.h"
#include "geometry.h"
#include "iofiles.h"
#include "multigrid.h"
#include "noise.h"
#include "diffuser.h"

/*---------------------------------------------------------------------------*/
static int simulate_statics(field_geometry *geometry,
                            simulation_condition *condition,
                            simulation_iofiles *iofiles,
                            polymer_blend *blend,density_field *density,
                            external_force *force,
                            field_fftw_plan *plan,
                            chemical_potential *potential,
                            multigrid_condition *mgcondition);
static int simulate_dynamics(field_geometry *geometry,
                             simulation_condition *condition,
                             simulation_iofiles *iofiles,
                             polymer_blend *blend,density_field *density,
                             external_force *force,
                             field_fftw_plan *plan,
                             chemical_potential *potential,
                             multigrid_condition *mgcondition);
static void write_data_to_file(simulation_condition *condition,
                               simulation_iofiles *iofiles,int n,
                               double fe, double fe_longrange,double fe_local,
                               double fe_gradient,density_field *density,
                               chemical_potential *potential);
static void resize_lattice(density_field *density,field_fftw_plan *plan);
static void calculate_free_energy_all(density_field *density,
                                      field_fftw_plan *plan,
                                      double *fe,double *fe_longrange,
                                      double *fe_local,double *fe_gradient);
static int filename_is_valid(char *filename);

/*---------------------------------------------------------------------------*/
int simulate(field_geometry *geometry,simulation_condition *condition,
             simulation_iofiles *iofiles,
             polymer_blend *blend,density_field *density,
             external_force *force,
             field_fftw_plan *plan,chemical_potential *potential,
             multigrid_condition *mgcondition)
{
    if(condition->interval == 0)
        condition->interval = condition->iteration_max;

    fprintf(stderr,"%s: %d\n",((geometry->dimension == 1) ? "dimension" : "dimensions"),
            geometry->dimension);
    fprintf(stderr,"simulation type: %s\n",(condition->dynamics_simulation ? "dynamics" : "statics"));
    fprintf(stderr,"lattice optimization: %s\n",(condition->optimize_lattice ? "true" : "false"));
    fprintf(stderr,"use multigrid solver: %s\n",(condition->use_multigrid_solver ? "true" : "false"));
    fprintf(stderr,"acceleration factor: %g\n",condition->omega);

    clear_chemical_potential(potential);

    if(condition->dynamics_simulation)
    {
        simulate_dynamics(geometry,condition,iofiles,blend,density,force,plan,
                          potential,mgcondition);
    }
    else
    {
        simulate_statics(geometry,condition,iofiles,blend,density,force,plan,
                         potential,mgcondition);
    }

    return 0;
}

/*---------------------------------------------------------------------------*/
int simulate_statics(field_geometry *geometry,
                     simulation_condition *condition,
                     simulation_iofiles *iofiles,
                     polymer_blend *blend,density_field *density,
                     external_force *force,
                     field_fftw_plan *plan,
                     chemical_potential *potential,
                     multigrid_condition *mgcondition)
{
    int i;
    int n,n_tridiag;
    double fe,fe_longrange,fe_local,fe_gradient,fe_old;
    double err;
    double omega;
    tridiag_matrix *tridiag;
    double *u;
    multigrid_field *multigrid;

    omega = condition->omega;

    calculate_free_energy_all(density,plan,
                              &fe,&fe_longrange,&fe_local,&fe_gradient);
    fprintf(stderr,"free energy / unit volume: %g\n",fe);
    fe_old = fe;
    print_lattice(stderr,geometry);


    n = lattice_point_number(density->geometry);
    n_tridiag = maximum_number_of_division_in_one_axis(density->geometry);

    tridiag = allocate_tridiag_matrix(n_tridiag);
    u = allocate_vector(n_tridiag);

    multigrid = allocate_multigrid_field(density->geometry);

    err = 0;
    n = 0;

    calculate_chemical_potential(blend,density,potential,plan);

    for(i = 0;i < condition->iteration_max;i++)
    {
        evolve_density_adi(density,potential,force,omega,condition->noise,
                           plan,mgcondition,multigrid,condition->use_multigrid_solver,
                           tridiag,u);

        if(((i % condition->interval) == 0) && (i > 0))
        {
            calculate_free_energy_all(density,plan,
                                      &fe,&fe_longrange,&fe_local,&fe_gradient);

            calculate_phi_from_psi(density);
            write_data_to_file(condition,iofiles,
                               n,fe,fe_longrange,
                               fe_local,fe_gradient,
                               density,potential);
            n++;
            err = fabs(fe - fe_old);

            fe_old = fe;

            if(condition->optimize_lattice)
                resize_lattice(density,plan);
        }

        if(((i % condition->interval) == 0) && (i > 0) &&
           (err < condition->error_tolerance))
            break;
    }


    fprintf(stderr,"iteration: %d\n",i);
    fprintf(stderr,"error: %g\n",err);

    calculate_chemical_potential(blend,density,potential,plan);
    calculate_free_energy_all(density,plan,
                              &fe,&fe_longrange,&fe_local,&fe_gradient);
    fprintf(stderr,"free energy / unit volume: %g\n",fe);
    print_lattice(stderr,geometry);


    calculate_phi_from_psi(density);
    write_data_to_file(condition,iofiles,
                       n,fe,fe_longrange,fe_local,fe_gradient,
                       density,potential);

    free_multigrid_field(multigrid);
    free_tridiag_matrix(tridiag);
    free_vector(u);

    return 0;
}

/*---------------------------------------------------------------------------*/
int simulate_dynamics(field_geometry *geometry,
                      simulation_condition *condition,
                      simulation_iofiles *iofiles,
                      polymer_blend *blend,density_field *density,
                      external_force *force,
                      field_fftw_plan *plan,
                      chemical_potential *potential,
                      multigrid_condition *mgcondition)
{
    int i;
    int n,n_tridiag;
    double fe,fe_longrange,fe_local,fe_gradient;
    double omega;
    tridiag_matrix *tridiag;
    multigrid_field *multigrid;
    double *u;

    omega = condition->omega;

    calculate_free_energy_all(density,plan,
                              &fe,&fe_longrange,&fe_local,&fe_gradient);
    fprintf(stderr,"free energy / unit volume: %g\n",fe);
    print_lattice(stderr,geometry);


    n = lattice_point_number(density->geometry);
    n_tridiag = maximum_number_of_division_in_one_axis(density->geometry);

    multigrid = allocate_multigrid_field(density->geometry);

    tridiag = allocate_tridiag_matrix(n_tridiag);
    u = allocate_vector(n_tridiag);

    n = 0;


    calculate_chemical_potential(blend,density,potential,plan);

    calculate_phi_from_psi(density);
    for(i = 0;i < condition->iteration_max;i++)
    {
        diffuse_density_adi(density,potential,force,omega,condition->noise,
                            plan,NULL,NULL,FALSE,
                            tridiag,u);

        correct_negative_density_field_with_threshold(density,
                                                      condition->phi_min);

        if(((i % condition->interval) == 0) && (i > 0))
        {
            calculate_free_energy_all(density,plan,
                                      &fe,&fe_longrange,&fe_local,&fe_gradient);

            write_data_to_file(condition,iofiles,
                               n,fe,fe_longrange,
                               fe_local,fe_gradient,
                               density,potential);
            n++;

            if(condition->optimize_lattice)
                resize_lattice(density,plan);
        }
    }


    fprintf(stderr,"iteration: %d\n",i);

    calculate_chemical_potential(blend,density,potential,plan);


    calculate_free_energy_all(density,plan,
                              &fe,&fe_longrange,&fe_local,&fe_gradient);
    fprintf(stderr,"free energy / unit volume: %g\n",fe);
    print_lattice(stderr,geometry);


    write_data_to_file(condition,iofiles,
                       n,fe,fe_longrange,fe_local,fe_gradient,
                       density,potential);

    free_multigrid_field(multigrid);

    free_tridiag_matrix(tridiag);
    free_vector(u);

    return 0;
}

/*---------------------------------------------------------------------------*/
void write_data_to_file(simulation_condition *condition,
                        simulation_iofiles *iofiles,int n,
                        double fe, double fe_longrange,double fe_local,
                        double fe_gradient,density_field *density,
                        chemical_potential *potential)
{
    static char filename[FILENAME_BUFFERSIZE];

    if(condition->save_density_sequential)
    {
        sprintf(filename,iofiles->density_template,n);
        write_density_plot(filename,density);
    }
    else if(filename_is_valid(iofiles->density_output))
    {
        sprintf(filename,"%s~",iofiles->density_output);
        write_density_plot(filename,density);
        rename(filename,iofiles->density_output);
    }

    if(condition->save_chemical_potential_sequential)
    {
        sprintf(filename,iofiles->chemical_potential_template,n);
        write_chemical_potential_plot(filename,potential);
    }
    else if(filename_is_valid(iofiles->chemical_potential_output))
    {
        sprintf(filename,"%s~",iofiles->chemical_potential_output);
        write_chemical_potential_plot(filename,potential);
        rename(filename,iofiles->chemical_potential_output);
    }

    if(condition->save_psi_sequential)
    {
        sprintf(filename,iofiles->psi_template,n);
        write_psi_field_text(filename,density);
    }
    else if(filename_is_valid(iofiles->psi_output))
    {
        sprintf(filename,"%s~",iofiles->psi_output);
        write_psi_field_text(filename,density);
        rename(filename,iofiles->psi_output);
    }

    if(condition->save_free_energy_sequential)
    {
        sprintf(filename,iofiles->free_energy_template,n);
        write_free_energy_text(filename,fe,fe_longrange,fe_local,fe_gradient);
    }
    else if(filename_is_valid(iofiles->free_energy_output))
    {
        sprintf(filename,"%s~",iofiles->free_energy_output);
        write_free_energy_text(filename,fe,fe_longrange,fe_local,fe_gradient);
        rename(filename,iofiles->free_energy_output);
    }

    if(condition->save_geometry_sequential)
    {
        sprintf(filename,iofiles->geometry_template,n);
        write_geometry_text(filename,density->geometry);
    }
    else if(filename_is_valid(iofiles->geometry_output))
    {
        sprintf(filename,"%s~",iofiles->geometry_output);
        write_geometry_text(filename,density->geometry);
        rename(filename,iofiles->geometry_output);
    }

    if(condition->save_dx_sequential)
    {
        sprintf(filename,iofiles->dx_template,n);
        write_field_dx(filename,density,potential);
    }
    else if(filename_is_valid(iofiles->dx_output))
    {
        sprintf(filename,"%s~",iofiles->dx_output);
        write_field_dx(filename,density,potential);
        rename(filename,iofiles->dx_output);
    }
}

/*---------------------------------------------------------------------------*/
void resize_lattice(density_field *density,field_fftw_plan *plan)
{
    double df,df2,lx0;
    field_geometry *geometry;

    geometry = density->geometry;

    df = calculate_free_energy_differential(density,plan);
    df2 = calculate_free_energy_differential2(density,plan);
    lx0 = geometry->lx;
    geometry->lx += df / df2;
    geometry->ly *= geometry->lx / lx0;
    geometry->lz *= geometry->lx / lx0;
    calculate_q2(geometry);
    calculate_q2i(geometry);
}

/*---------------------------------------------------------------------------*/
void calculate_free_energy_all(density_field *density,
                               field_fftw_plan *plan,
                               double *fe,double *fe_longrange,
                               double *fe_local,double *fe_gradient)
{
    *fe = calculate_free_energy(density,plan);
    *fe_longrange = calculate_free_energy_longrange(density,plan);
    *fe_local = calculate_free_energy_local(density);
    *fe_gradient = calculate_free_energy_gradient(density,plan);
}

/*---------------------------------------------------------------------------*/
int filename_is_valid(char *filename)
{
    if((filename == NULL) || (strlen(filename) == 0))
    {
        return FALSE;
    }
    else
    {
        return TRUE;
    }
}

/*---------------------------------------------------------------------------*/
