/*---------------------------------------------------------------------------*/

/*!  \file  tridiag.h
     \brief  solver for tridiagonal and cyclic tridiagonal matrices  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2005 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#ifndef INCLUDE_TRIDIAG_H
#define INCLUDE_TRIDIAG_H

/*---------------------------------------------------------------------------*/
/*!  \brief  tridiagonal matrix \a A.  */
typedef struct tridiag_matrix
{
        /*!  \brief  the diagonal elements of \a A.  */
    double *d;
        /*!  \brief  the lower subdiagonal elements of \a A.  */
    double *dl;
        /*!  \brief  the upper subdiagonal elements of \a A.  */
    double *du;
} tridiag_matrix;

/*---------------------------------------------------------------------------*/
/*!  \brief  allocate tridiagonal (or cyclic tridiagonal) matrix \a A.
     \param  n  number of elements for \a d, \a dl and \a du.  */
tridiag_matrix *allocate_tridiag_matrix(int n);

/*!  \brief  free tridiagonal matrix.
     \param  tridiag  tridiagonal matrix \a A to free.  */
void free_tridiag_matrix(tridiag_matrix *tridiag);

/*!  \brief  solve \a A \a x = \a b where \a A is the tridiagonal matrix.
     \param  n  dimension of matrix \a A.
     \param  n_rhs  the number of right hand sides, \b.
     \param  n_stride  the size of strides between the elements of \a b.
     \param  tridiag  tridiagonal matrix \a A (the diagonal elements \a d will be destroyed by the function).
     \param  b  input/output vector (overwritten to \a x).
     */
void solve_tridiag(int n,int n_rhs,int n_stride,
                   tridiag_matrix *tridiag,double *b);
/*!  \brief  solve \a A \a x = \a b with Sherman-Morrison formula where \a A is the cyclic tridiagonal matrix.
     \param  n  dimension of matrix \a A.
     \param  n_rhs  the number of right hand sides, \b.
     \param  n_stride  the size of strides between the elements of \a b.
     \param  tridiag  tridiagonal matrix \a A (the diagonal elements \a d will be destroyed by the function).
     \param  b  input/output vector (overwritten to \a x).
     \param  u  work array (size \a n).
     */
void solve_cyclic_tridiag(int n,int n_rhs,int n_stride,
                          tridiag_matrix *tridiag,double *b,double *u);

/*---------------------------------------------------------------------------*/
#endif

/*---------------------------------------------------------------------------*/
