/*---------------------------------------------------------------------------*/

/*!  \file  dxread.c
     \brief  read field data from DX file generated by drops.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2007 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include "xmalloc.h"
#include "dxread.h"
#include "field.h"

#define STRING_BUFFERSIZE 0x100

/*---------------------------------------------------------------------------*/
typedef enum field_type
{
    phi_field, mu_field
} field_type;

typedef enum endian_type
{
    lsb, msb
} endian_type;

typedef struct field_information
{
    int index;
    field_type field;
    endian_type endian;
    int size;
    int offset;
    struct field_information *next;
} field_information;

/*---------------------------------------------------------------------------*/
static int read_field_dx_geometry(FILE *fp,field_geometry *geometry);
static field_information *read_field_dx_field_information(FILE *fp,field_geometry *geometry);
static int skip_until_binary_data_block(FILE *fp);
static int check_field_information_consistency(field_information *information);
static void free_field_information_list(field_information *information);
static void read_field_dx_field_data_binary(FILE *fp,void *u,
                                            field_geometry *geometry,
                                            field_information *information,
                                            long offset_binary);

/*---------------------------------------------------------------------------*/
int read_field_dx(char *filename,field_geometry *geometry,
                  double ***phi,double ***mu,int *n)
{
    FILE *fp;
    field_information *information;
    field_information *information_phi,*information_mu;
    long offset_binary;
    double **phi_double,**mu_double;
    float **phi_float,**mu_float;

    if((filename == NULL) || (strlen(filename) == 0))
        return -1;

    fp = fopen(filename,"rb");


    if(read_field_dx_geometry(fp,geometry) != 0)
    {
        fprintf(stderr,"error: cannot read geometrical information from `%s'.\n",
                filename);

        fclose(fp);
        return -1;
    }


    information_phi = NULL;
    information_mu = NULL;
    while((information = read_field_dx_field_information(fp,geometry)) != NULL)
    {
        switch(information->field)
        {
            case phi_field:
                information->next = information_phi;
                information_phi = information;
                break;

            case mu_field:
                information->next = information_mu;
                information_mu = information;
                break;

            default:
                break;
        }
    }

    if((check_field_information_consistency(information_phi) != 0)
       || (check_field_information_consistency(information_mu) != 0))
    {
        fprintf(stderr,"error: field data in `%s' are not consistent.\n",
                filename);

        fclose(fp);
        return -1;
    }

    if(skip_until_binary_data_block(fp) != 0)
    {
        fprintf(stderr,"error: cannot find the binary data block in `%s'.\n",
                filename);

        fclose(fp);
        return -1;
    }

    offset_binary = ftell(fp);


    phi_double = allocate_field_double(geometry,information_phi->index + 1);
    mu_double = allocate_field_double(geometry,information_phi->index + 1);
    if(information_phi->size == 4)
    {
        phi_float = allocate_field_float(geometry,information_phi->index + 1);
        mu_float = allocate_field_float(geometry,information_phi->index + 1);
    }

    information = information_phi;
    while(information != NULL)
    {
        if(information_phi->size == 4)
        {
            read_field_dx_field_data_binary(fp,phi_float[information->index],
                                            geometry,information,offset_binary);
        }
        else
        {
            read_field_dx_field_data_binary(fp,phi_double[information->index],
                                            geometry,information,offset_binary);
        }

        information = information->next;
    }

    information = information_mu;
    while(information != NULL)
    {
        if(information_mu->size == 4)
        {
            read_field_dx_field_data_binary(fp,mu_float[information->index],
                                            geometry,information,offset_binary);
        }
        else
        {
            read_field_dx_field_data_binary(fp,mu_double[information->index],
                                            geometry,information,offset_binary);
        }

        information = information->next;
    }


    fclose(fp);



    if(information_phi->size == 4)
    {
        convert_field_float_to_double(phi_float,phi_double,
                                      geometry,information_phi->index + 1);
        convert_field_float_to_double(mu_float,mu_double,
                                      geometry,information_mu->index + 1);

        free_field_float(phi_float);
        free_field_float(mu_float);
    }

    *phi = phi_double;
    *mu = mu_double;

    *n = information_phi->index + 1;

    free_field_information_list(information_phi);
    free_field_information_list(information_mu);

    return 0;
}

/*---------------------------------------------------------------------------*/
int read_field_dx_geometry(FILE *fp,field_geometry *geometry)
{
    static char buf[STRING_BUFFERSIZE];
    int dimension,n[3];
    double l[3];
    int i,j;
    double d;

        /*  skip object information about "positions"  */
    fscanf(fp,"object \"positions\" class gridpositions counts");

        /*  read field geometry information  */
    dimension = 0;
    while(fscanf(fp,"%s",buf) > 0)
    {
        if(strcmp(buf,"origin") == 0)
        {
            break;
        }
        else if(dimension == 3)
            return -1;

        sscanf(buf,"%d",&(n[dimension]));
        dimension++;
    }


    for(i = 0;i < dimension;i++)
    {
        if(!(fscanf(fp,"%s",buf) > 0))
            return -1;

        sscanf(buf,"%lf",&d);

        if(d != 0.0)
            return -1;
    }

    for(i = 0;i < dimension;i++)
    {
        if(!(fscanf(fp,"%s",buf) > 0))
            return -1;

        if(strcmp(buf,"delta") != 0)
            return -1;

        for(j = 0;j < dimension;j++)
        {
            if(!(fscanf(fp,"%s",buf) > 0))
                return -1;

            if(j == i)
            {
                sscanf(buf,"%lf",&(l[i]));

                l[i] *= n[i];
            }
            else
            {
                sscanf(buf,"%lf",&d);

                if(d != 0.0)
                    return -1;
            }
        }
    }


        /*  skip object information about "connections"  */
    while(fscanf(fp,"\n") > 0)
        ;
    fscanf(fp,"object \"connections\" class gridconnections counts");

        /*  verify field geometry information  */
    for(i = 0;i < dimension;i++)
    {
        if(!(fscanf(fp,"%s",buf) > 0))
            return -1;

        sscanf(buf,"%d",&j);

        if(j != n[i])
            return -1;
    }


        /*  set field geometry  */
    geometry->dimension = dimension;
    for(i = 0;i < dimension;i++)
    {
        geometry->n[i] = n[i];
        geometry->l[i] = l[i];
    }
    for(i = dimension;i < 3;i++)
    {
        geometry->n[i] = 1;
        geometry->l[i] = 1.0;
    }

    return 0;
}

/*---------------------------------------------------------------------------*/
field_information *read_field_dx_field_information(FILE *fp,field_geometry *geometry)
{
    static char buf[STRING_BUFFERSIZE];
    int index,count,offset,size;
    field_information *information;
    field_type field;
    endian_type endian;

        /*  read object information about "data_phi" or "data_mu" */
    while(fscanf(fp,"\n") > 0)
        ;
    fscanf(fp,"object");

    if(!(fscanf(fp,"%s",buf) > 0))
        return NULL;
    if(sscanf(buf,"\"data_phi%d\"",&index) > 0)
    {
        field = phi_field;
/*         printf("# phi %d\n",n); */
    }
    else if(sscanf(buf,"\"data_mu%d\"",&index) > 0)
    {
        field = mu_field;
/*         printf("# mu %d\n",n); */
    }
    else
    {
        return NULL;
    }

    while(fscanf(fp,"\n") > 0)
        ;
    fscanf(fp,"class array type");


    if(!(fscanf(fp,"%s",buf) > 0))
        return NULL;
    if(strcmp(buf,"float") == 0)
    {
        size = 4;
/*         printf("# float (4)\n"); */
    }
    else if(strcmp(buf,"double") == 0)
    {
        size = 8;
/*         printf("# double (8)\n",n); */
    }
    else
    {
        return NULL;
    }

    if(!(fscanf(fp,"%s",buf) > 0))
        return NULL;

    if(!(fscanf(fp,"%s",buf) > 0))
        return NULL;
    if(sscanf(buf,"%d",&count) == 0)
        return NULL;
/*     printf("# %d\n",count); */
    if(count != lattice_point_number(geometry))
        return NULL;

    if(!(fscanf(fp,"%s",buf) > 0))
        return NULL;
    if(strcmp(buf,"lsb") == 0)
    {
/*         printf("# lsb\n"); */
        endian = lsb;
    }
    else if(strcmp(buf,"msb") == 0)
    {
/*         printf("# msb\n"); */
        endian = msb;
    }
    else
    {
        return NULL;
    }

    if(!(fscanf(fp,"%s",buf) > 0))
        return NULL;
    if(!(fscanf(fp,"%s",buf) > 0))
        return NULL;

    if(!(fscanf(fp,"%s",buf) > 0))
        return NULL;
    if(sscanf(buf,"%d",&offset) == 0)
        return NULL;
/*     printf("# %d\n",offset); */


        /*  set field information  */
    information = xmalloc(sizeof(field_information));

    information->index = index;
    information->field = field;
    information->endian = endian;
    information->size = size;
    information->offset = offset;
    information->next = NULL;

    return information;
}

/*---------------------------------------------------------------------------*/
int skip_until_binary_data_block(FILE *fp)
{
    static char buf[STRING_BUFFERSIZE];

    while(fscanf(fp,"%s",buf) > 0)
    {
        if(strcmp(buf,"end") == 0)
        {
            fscanf(fp,"\n");
            return 0;
        }
    }

    return -1;
}

/*---------------------------------------------------------------------------*/
void free_field_information_list(field_information *information)
{
    if(information->next == NULL)
    {
        free(information);
    }
    else
    {
        free_field_information_list(information->next);
        free(information);
    }
}

/*---------------------------------------------------------------------------*/
int check_field_information_consistency(field_information *information)
{
    field_information *root;
    int i;

    root = information;
    i = information->index;

    information = information->next;
    while(information != NULL)
    {
        if((root->endian != information->endian)
           || (root->size != information->size))
            return -1;

        i--;
        if(information->index != i)
            return -1;

        information = information->next;
    }

    return 0;
}

/*---------------------------------------------------------------------------*/
void read_field_dx_field_data_binary(FILE *fp,void *u,
                                     field_geometry *geometry,
                                     field_information *information,
                                     long offset_binary)
{
    fseek(fp,offset_binary + information->offset,SEEK_SET);

    fread(u,information->size,lattice_point_number(geometry),fp);
}

/*---------------------------------------------------------------------------*/
